from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
from xoa_driver.internals.core.commands import (
    P4G_CLIENT_RANGE,
    P4G_SERVER_RANGE,
    P4G_IP_DS_TYPE,
    P4G_IP_DS_VALUE,
    P4G_IP_DS_MASK,
    P4G_IP_DS_MINMAX,
    P4G_IP_DS_STEP,
    P4G_IP_VERSION,
    P4G_IPV6_CLIENT_RANGE,
    P4G_IPV6_SERVER_RANGE,
    P4G_IPV6_TRAFFIC_CLASS,
    P4G_IPV6_FLOW_LABEL,
    P4G_NAT,
)


class GIPv4L3:
    """L3 IPv4"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.client_range = P4G_CLIENT_RANGE(conn, module_id, port_id, group_idx)
        """
        Representation P4G_CLIENT_RANGE
        """
        self.server_range = P4G_SERVER_RANGE(conn, module_id, port_id, group_idx)
        """
        Representation P4G_SERVER_RANGE
        """


class GIPv6L3:
    """L3 IPv6"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.client_range = P4G_IPV6_CLIENT_RANGE(conn, module_id, port_id, group_idx)
        """
        Representation P4G_IPV6_CLIENT_RANGE
        """
        self.server_range = P4G_IPV6_SERVER_RANGE(conn, module_id, port_id, group_idx)
        """
        Representation P4G_IPV6_SERVER_RANGE
        """
        self.traffic_class = P4G_IPV6_TRAFFIC_CLASS(conn, module_id, port_id, group_idx)
        """
        Representation P4G_IPV6_TRAFFIC_CLASS
        """
        self.flow_label = P4G_IPV6_FLOW_LABEL(conn, module_id, port_id, group_idx)
        """
        Representation P4G_IPV6_FLOW_LABEL
        """


class GDifferentialServiceL3:
    """L3 Differential Service"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.type = P4G_IP_DS_TYPE(conn, module_id, port_id, group_idx)
        """
        Representation P4G_IP_DS_TYPE
        """
        self.value = P4G_IP_DS_VALUE(conn, module_id, port_id, group_idx)
        """
        Representation P4G_IP_DS_VALUE
        """
        self.mask = P4G_IP_DS_MASK(conn, module_id, port_id, group_idx)
        """
        Representation P4G_IP_DS_MASK
        """
        self.range_limits = P4G_IP_DS_MINMAX(conn, module_id, port_id, group_idx)
        """
        Representation P4G_IP_DS_MINMAX
        """
        self.step = P4G_IP_DS_STEP(conn, module_id, port_id, group_idx)
        """
        Representation P4G_IP_DS_STEP
        """


class GL3:
    """L3 Configurations"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.ip_version = P4G_IP_VERSION(conn, module_id, port_id, group_idx)
        """
        Representation P4G_IP_VERSION
        """
        self.ipv4 = GIPv4L3(conn, module_id, port_id, group_idx)
        """L3 IPv4 Configurations"""
        self.ipv6 = GIPv6L3(conn, module_id, port_id, group_idx)
        """L3 IPv6 Configurations"""
        self.diffserv = GDifferentialServiceL3(conn, module_id, port_id, group_idx)
        """L3 IPv6 DS Configurations"""
        self.nat = P4G_NAT(conn, module_id, port_id, group_idx)  # TODO: maybe to add extra logic for this command based on the description
        """
        Representation P4G_NAT
        """



from typing import (
    NamedTuple,
    Type,
    TypeVar,
)
from xoa_driver.internals.core.commands import P_CAPABILITIES

T = TypeVar("T", bound="CapID")


class CapID(NamedTuple):
    """ID generated based on the port capability"""

    can_set_autoneg: int
    can_eee: int
    can_hw_reg_access: int
    can_tcvr_mii_reg_access: int
    can_adv_phy_man: int
    can_mdi_mdix: int
    can_dyn_traffic_change: int
    can_pcs_pma_config: int
    can_fec: int
    can_fec_stats: int
    can_tx_eq: int
    can_rx_retune: int
    can_manipulate_preamble: int
    can_set_link_train: int
    can_link_flap: int
    can_auto_neg_base_r: int
    can_pma_error_pulse: int

    @classmethod
    def create_from_capabilities(cls: Type[T], cap: P_CAPABILITIES.GetDataAttr) -> T:
        return cls(
            int(cap.can_set_autoneg),
            int(cap.can_eee),
            int(cap.can_hw_reg_access),
            int(cap.can_tcvr_mii_reg_access),
            int(cap.can_adv_phy_man),
            int(cap.can_mdi_mdix),
            int(cap.can_dyn_traffic_change),
            int(cap.can_pcs_pma_config),
            int(cap.can_fec),
            int(cap.can_fec_stats),
            int(cap.can_tx_eq),
            int(cap.can_rx_retune),
            int(cap.can_manipulate_preamble),
            int(cap.can_set_link_train),
            int(cap.can_link_flap),
            int(cap.can_auto_neg_base_r),
            int(cap.can_pma_error_pulse),
        )

    def to_int(self) -> int:
        return int("".join(str(ele) for ele in self), 2)

    def __eq__(self, other: "CapID") -> bool:
        if isinstance(other, CapID):
            return self.to_int() == other.to_int()
        return False

    def __ne__(self, other: "CapID") -> bool:
        return not self.__eq__(other)



import asyncio
from typing import (
    TYPE_CHECKING,
    Tuple
)
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
from xoa_driver.internals.core.transporter import funcs
from xoa_driver.internals.core.transporter.exceptions import BadStatus
from xoa_driver.internals import exceptions
from xoa_driver import enums
from xoa_driver.internals.core.commands import (
    C_LOGON,
    C_LOGOFF,
    C_OWNER,
    C_TIMEOUT,
    C_KEEPALIVE,
    C_INDICES,
    C_STATSESSION,
)


class TesterSession:
    """
    Establishing communication session through connection with ester unit.
    """

    __slots__ = ("_conn", "owner_name", "pwd", "timeout", "keepalive")

    def __init__(self, conn: "itf.IConnection", owner_name: str, password: str = "xena", timeout_seconds: int = 130, keepalive: bool = False) -> None:
        self._conn = conn
        self.owner_name = owner_name
        self.pwd = password
        self.timeout = timeout_seconds
        self.keepalive = keepalive

    def __await__(self):
        return self.__setup().__await__()

    async def __setup(self) -> "TesterSession":
        try:
            await funcs.apply(
                C_LOGON(self._conn).set(self.pwd),
                C_OWNER(self._conn).set(self.owner_name),
                C_TIMEOUT(self._conn).set(self.timeout),
            )
        except BadStatus as e:
            if e.response.class_name == "C_LOGON" and e.response.command_status == enums.CommandStatus.BADVALUE:
                raise exceptions.WrongTesterPasswordError(self.pwd) from None
            raise e
        if self.keepalive:
            asyncio.create_task(self.__do_keepalive(), name="keepalive")  # TODO: need to handle this task on interruption
        return self

    async def __do_keepalive(self) -> None:
        while self._conn.is_connected:
            await C_KEEPALIVE(self._conn).get()
            await asyncio.sleep(max(0, self.timeout - 5))

    @property
    def is_online(self) -> bool:
        """
        Check if connection is still active.
        """

        return self._conn.is_connected

    async def logoff(self) -> None:
        """
        Gracefully log off from the tester.
        """

        await C_LOGOFF(self._conn).set()
        self._conn.close()

    async def sessions_info(self) -> Tuple[C_STATSESSION.GetDataAttr, ...]:
        """
        Return information about all active sessions on the tester.
        """

        sessions = await C_INDICES(self._conn).get()
        session_ids = sessions.session_ids
        query_sessions = [
            C_STATSESSION(self._conn, session_xindex=sid).get()
            for sid in session_ids
        ]
        return tuple(await funcs.apply(*query_sessions))



import asyncio
import functools
from typing import (
    TYPE_CHECKING,
    Optional,
)
from typing_extensions import Self
from xoa_driver.internals.core.commands import (
    M_STATUS,
    M_UPGRADE,
    M_UPGRADEPROGRESS,
    M_TIMESYNC,
    M_CFPTYPE,
    M_COMMENT,
    # M_TIMEADJUSTMENT,
    M_CAPABILITIES,
    M_MEDIASUPPORT,
    M_FPGAREIMAGE,
    M_MULTIUSER,
    M_CFPCONFIGEXT,
    M_CFPCONFIG,
    M_CLOCKPPB,
    M_SMAINPUT,
    M_SMAOUTPUT,
    M_SMASTATUS,
    M_NAME,
    M_REVISION,
    M_MEDIA,
    M_CLOCKSYNCSTATUS,
    M_TXCLOCKSOURCE_NEW,
    M_TXCLOCKSTATUS_NEW,
    M_TXCLOCKFILTER_NEW,
)
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf

from xoa_driver.internals.utils import attributes as utils
from xoa_driver.internals.utils import ports_manager as pm
from xoa_driver.internals.state_storage import modules_state

from .. import base_module as bm
from .. import __interfaces as m_itf


class TXClock:
    """Advanced timing clock"""
    def __init__(self, conn: "itf.IConnection", module_id: int) -> None:
        self.source = M_TXCLOCKSOURCE_NEW(conn, module_id)
        """The source that drives the TX clock rate of the ports on the test module.
        Representation of M_TXCLOCKSOURCE_NEW
        """

        self.status = M_TXCLOCKSTATUS_NEW(conn, module_id)
        """TX clock status of the test module.
        Representation of M_TXCLOCKSTATUS_NEW
        """

        self.filter = M_TXCLOCKFILTER_NEW(conn, module_id)
        """Loop bandwidth on the TX clock filter of the test module.
        Representation of M_TXCLOCKFILTER_NEW
        """


class SMA:
    """SMA connector"""
    def __init__(self, conn: "itf.IConnection", module_id: int) -> None:
        self.input = M_SMAINPUT(conn, module_id)
        """SMA input of the test module.
        Representation of M_SMAINPUT
        """

        self.output = M_SMAOUTPUT(conn, module_id)
        """SMA output of the test module.
        Representation of M_SMAOUTPUT
        """

        self.status = M_SMASTATUS(conn, module_id)
        """SMA input status of the test module.
        Representation of M_SMASTATUS
        """


class AdvancedTiming:
    """Advanced Timing config and control"""
    def __init__(self, conn: "itf.IConnection", module_id: int) -> None:
        self.clock_tx = TXClock(conn, module_id)
        """Advanced timing clock config and status
        """

        self.sma = SMA(conn, module_id)
        """SMA connector
        """


class CFP:
    """Test module CFP"""
    def __init__(self, conn: "itf.IConnection", module_id: int) -> None:
        self.type = M_CFPTYPE(conn, module_id)
        """The transceiver's CFP type currently inserted.
        Representation of M_CFPTYPE
        """

        self.config = M_CFPCONFIG(conn, module_id)
        """The CFP configuration of the test module.
        Representation of M_CFPCONFIG
        """

        self.config_extended = M_CFPCONFIGEXT(conn, module_id)
        """The CFP configuration of the test module.
        Representation of M_CFPCONFIGEXT
        """


class MTiming:
    """Test module timing and clock configuration"""
    def __init__(self, conn: "itf.IConnection", module_id: int) -> None:
        self.source = M_TIMESYNC(conn, module_id)
        """Timing source of the test module.
        Representation of M_TIMESYNC
        """

        self.clock_local_adjust = M_CLOCKPPB(conn, module_id)
        """Time adjustment controlling of the local clock of the test module, which drives the TX rate of the test ports.
        Representation of M_CLOCKPPB
        """

        self.clock_sync_status = M_CLOCKSYNCSTATUS(conn, module_id)
        """Test module's clock sync status.
        Representation of M_CLOCKSYNCSTATUS
        """


class MUpgrade:
    """Test module upgrade"""
    def __init__(self, conn: "itf.IConnection", module_id: int) -> None:
        self.start = M_UPGRADE(conn, module_id)
        """Start the upgrade progress of the test module.
        Representation of M_UPGRADE
        """

        self.progress = M_UPGRADEPROGRESS(conn, module_id)
        """Upgrade progress status of the test module.
        Representation of M_UPGRADEPROGRESS
        """

        self.reload_image = M_FPGAREIMAGE(conn, module_id)
        """Reload the FPGA image of the test module.
        Representation of M_FPGAREIMAGE
        """


class ModuleL23(bm.BaseModule["modules_state.ModuleL23LocalState"]):
    """
    Representation of a L23 test module on a physical tester.
    """
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)

        self._local_states = modules_state.ModuleL23LocalState()

        self.name = M_NAME(conn, self.module_id)
        """Test module's name.
        Representation of M_NAME
        """

        self.comment = M_COMMENT(conn, self.module_id)
        """Test module's description.
        Representation of M_COMMENT
        """

        self.status = M_STATUS(conn, self.module_id)
        """Test module's status.
        Representation of M_STATUS
        """

        self.media = M_MEDIA(conn, self.module_id)
        """Test module's media configuration.
        Representation of M_MEDIA
        """

        self.available_speeds = M_MEDIASUPPORT(conn, self.module_id)
        """Test module's available speeds.
        Representation of M_MEDIASUPPORT
        """

        self.revision = M_REVISION(conn, self.module_id)
        """Test module's model P/N name.
        Representation of M_REVISION
        """

        self.multiuser = M_MULTIUSER(conn, self.module_id)
        """If multiple users are allowed to control the same test module.
        Representation of M_MULTIUSER
        """

        self.capabilities = M_CAPABILITIES(conn, self.module_id)
        """Test module's capabilities.
        Representation of M_CAPABILITIES
        """

        self.timing = MTiming(conn, self.module_id)
        """Test module's timing configuration."""

        self.advanced_timing = AdvancedTiming(conn, self.module_id)
        """Test module's advanced timing ."""

        self.cfp = CFP(conn, self.module_id)
        """Test module's CFP """

        self.upgrade = MUpgrade(conn, self.module_id)
        """Test module's upgrade.

        :type: MUpgrade
        """

        self.ports: Optional[pm.PortsManager] = None
        """L23 port index manager of the test module."""

    @property
    def info(self) -> modules_state.ModuleL23LocalState:
        return self._local_states

    async def _setup(self) -> Self:
        if self.ports is None:
            raise NotImplementedError("Ports manager type are not defined.")  # Maybe can be better solution then this...
        await asyncio.gather(
            self._local_states.initiate(self),
            self.ports.fill()
        )
        self._local_states.register_subscriptions(self)
        return self

    on_cfp_type_change = functools.partialmethod(utils.on_event, M_CFPTYPE)
    """Register a callback to the event that the module's CFP type changes."""

    on_cfp_config_change = functools.partialmethod(utils.on_event, M_CFPCONFIG)
    """Register a callback to the event that the module's CFP configuration changes."""

    on_cfp_config_extended_change = functools.partialmethod(utils.on_event, M_CFPCONFIGEXT)
    """Register a callback to the event that the module's CFP configuration changes."""

    on_status_change = functools.partialmethod(utils.on_event, M_STATUS)
    """Register a callback to the event that the module's status changes."""

    on_revision_change = functools.partialmethod(utils.on_event, M_REVISION)
    """Register a callback to the event that the module's model type changes."""

    on_name_change = functools.partialmethod(utils.on_event, M_NAME)
    """Register a callback to the event that the module's name changes."""

    on_adv_timing_clock_tx_status_change = functools.partialmethod(utils.on_event, M_TXCLOCKSTATUS_NEW)
    """Register a callback to the event that the module's TX clock status changes."""

    on_adv_timing_clock_tx_source_change = functools.partialmethod(utils.on_event, M_TXCLOCKSOURCE_NEW)
    """Register a callback to the event that the module's clock that drives the port TX rates changes."""

    on_adv_timing_clock_tx_filter_change = functools.partialmethod(utils.on_event, M_TXCLOCKFILTER_NEW)
    """Register a callback to the event that the module's loop bandwidth on the TX clock filter changes."""

    on_adv_timing_sma_status_change = functools.partialmethod(utils.on_event, M_SMASTATUS)
    """Register a callback to the event that the module's SMA status changes."""

    on_adv_timing_sma_input_change = functools.partialmethod(utils.on_event, M_SMAINPUT)
    """Register a callback to the event that the module's SMA input function changes."""

    on_adv_timing_sma_output_change = functools.partialmethod(utils.on_event, M_SMAOUTPUT)
    """Register a callback to the event that the module's SMA output function changes."""

    on_media_support_change = functools.partialmethod(utils.on_event, M_MEDIASUPPORT)
    """Register a callback to the event that the module's supported media changes."""

    on_media_change = functools.partialmethod(utils.on_event, M_MEDIA)
    """Register a callback to the event that the module's media and available speeds change."""

    on_comment_change = functools.partialmethod(utils.on_event, M_COMMENT)
    """Register a callback to the event that the module's description changes."""

    on_timing_source_change = functools.partialmethod(utils.on_event, M_TIMESYNC)
    """Register a callback to the event that the module's timesync mode changes."""

    on_timing_clock_local_adjust_change = functools.partialmethod(utils.on_event, M_CLOCKPPB)
    """Register a callback to the event that the module's clock adjustment ppb changes."""



from typing import (
    Union,
    Any,
    Type,
    Iterable,
    Tuple,
    List
)

# region Converter Errors


class PaddingOddError(ValueError):
    def __init__(self, value: int) -> None:
        self.value = value
        self.msg = f"Parameter <'{self.value!r}'> must be an even number!"
        super().__init__(self.msg)


class PaddingToShortError(ValueError):
    def __init__(self, value: int, hex_string: str) -> None:
        self.value = value
        self.hex_string = hex_string
        self.msg = f"Parameter <'{self.value!r}'> is smaller than length of '{self.hex_string}' !"
        super().__init__(self.msg)


# endregion

# region DataTypes

class ExpectingBytesError(ValueError):
    def __init__(self, value: Any) -> None:
        self.value = value
        self.msg = f"Expected {self.value!r} to be an string of bytes."
        super().__init__(self.msg)


class NumberRangeError(ValueError):
    def __init__(self, value: int, max_val: int, min_val: int = 0) -> None:
        self.value = value
        self.min_val = min_val
        self.max_val = max_val
        self.msg = f"Expected({self.value!r}) to be not bigger than {self.max_val} and not smaller than {self.min_val}."
        super().__init__(self.msg)


class HexRangeError(ValueError):
    def __init__(self, value: str, max_val: int, min_val: int = 0) -> None:
        self.value = value
        self.min_val = min_val
        self.max_val = max_val
        self.msg = f"Expected {value!r} to be not bigger than {max_val} and not smaller than {self.min_val}."
        super().__init__(self.msg)


class InvalidMacAddressError(ValueError):
    def __init__(self, value: Union[int, str]) -> None:
        self.value = value
        self.msg = f"Invalid mac address {self.value!r}!"
        super().__init__(self.msg)


class StartWithError(ValueError):
    def __init__(self, value: str, *variations) -> None:
        self.value = value
        self.msg = f"Expected {self.value!r} to start with { ' or '.join(variations) }."
        super().__init__(self.msg)


class NotAllowedTypeError(TypeError):
    def __init__(self, value: Any, allowed_types: Iterable[Type]) -> None:
        self.value = value
        self.allowed_types = allowed_types
        self.msg = f"the type of <{type(self.value)} {self.value}> is not in allowed types { ' or '.join(allowed_types) }."
        super().__init__(self.msg)


class FixedLenghtError(ValueError):
    def __init__(self, value: Any, fix_length: int) -> None:
        self.value = value
        self.fix_length = fix_length
        self.msg = f"Data {self.value} failed to follow the specified fixed length {self.fix_length}."
        super().__init__(self.msg)


class ElementTypeError(NotImplementedError):
    def __init__(self, class_name: str) -> None:
        self.class_name = class_name
        self.msg = f"Ups, someone forgot to set a correct value for: <element_type> in <class {self.class_name}>"
        super().__init__(self.msg)

# endregion


# region Fields Errors

class ValueOutOfRangeError(ValueError):
    def __init__(self, value: Union[int, float, List[int], List[float]], climb: Tuple[Union[int, float], Union[int, float]]) -> None:
        self.value = value
        self.climb = climb
        self.msg = f"Expect value: <{value}> to be in range of {climb}!"
        super().__init__(self.msg)

# endregion



#: All available test module types.

from xoa_driver.internals.hli_v2.modules.modules_l23.module_l23_base import ModuleL23
from xoa_driver.internals.hli_v2.modules.modules_l23.family_d import (
    MOdin1G3S2PT,
    MOdin1G3S6P,
    MOdin1G3S6P_b,
    MOdin1G3S6PE,
)
from xoa_driver.internals.hli_v2.modules.modules_l23.family_e import (
    MOdin5G4S6PCU,
    MOdin10G3S2PCU,
    MOdin10G3S6PCU,
    MOdin10G5S6PCU,
    MOdin10G5S6PCU_b,
)
from xoa_driver.internals.hli_v2.modules.modules_l23.family_f import (
    MOdin10G1S2P,
    MOdin10G1S2P_b,
    MOdin10G1S2P_c,
    MOdin10G1S2P_d,
    MOdin10G1S2PT,
    MOdin10G1S6P,
    MOdin10G1S6P_b,
    MOdin10G1S12P,
)
from xoa_driver.internals.hli_v2.modules.modules_l23.family_g import (
    MLoki100G3S1P,
    MLoki100G3S1P_b,
    MLoki100G3S1PSE,
    MLoki100G3S1PB,
)
from xoa_driver.internals.hli_v2.modules.modules_l23.family_h import (
    MLoki100G5S1P,
    MOdin100G3S1P,
)
from xoa_driver.internals.hli_v2.modules.modules_l23.family_i import MLoki100G5S2P
from xoa_driver.internals.hli_v2.modules.modules_l23.family_j import MThor100G5S4P
from xoa_driver.internals.hli_v2.modules.modules_l23.family_k import MThor400G7S1P
from xoa_driver.internals.hli_v2.modules.modules_l23.family_l import (
    MThor400G7S1P_b,
    MThor400G7S1P_c,
    MThor400G7S1P_d,
)
from xoa_driver.internals.hli_v2.modules.modules_l23.family_l1 import (
    MFreya800G1S1P_a,
    MFreya800G1S1P_b,
    MFreya800G1S1POSFP_a,
    MFreya800G4S1P_a,
    MFreya800G4S1POSFP_a,
    MFreya800G4S1P_b,
)
from xoa_driver.internals.hli_v2.modules.modules_l23.family_m import MOdin1G3S6PT1RJ45
from xoa_driver.internals.hli_v2.modules.modules_l23.family_n import (
    MOdin40G2S2P,
    MOdin40G2S2PB,
)
from xoa_driver.internals.hli_v2.modules.modules_l23.family_combi import (
    MOdin10G4S2PCombi,
    MOdin10G4S2PCombi_b,
)
from xoa_driver.internals.hli_v2.modules.module_l23ve import ModuleL23VE
from xoa_driver.internals.hli_v2.modules.module_chimera import (
    ModuleChimera,
    MChi100G5S2P,
    MChi100G5S2P_b,
    MChi40G2S2P,
)
from xoa_driver.internals.hli_v2.modules.module_l47 import ModuleL47
from xoa_driver.internals.hli_v2.modules.module_l47ve import ModuleL47VE

import typing

GenericL23Module = typing.Union[
    "MOdin1G3S6P",
    "MOdin1G3S6P_b",
    "MOdin1G3S6PE",
    "MOdin1G3S2PT",
    "MOdin5G4S6PCU",
    "MOdin10G5S6PCU",
    "MOdin10G5S6PCU_b",
    "MOdin10G3S6PCU",
    "MOdin10G3S2PCU",
    "MOdin10G1S2P",
    "MOdin10G1S2P_b",
    "MOdin10G1S2P_c",
    "MOdin10G1S2P_d",
    "MOdin10G1S6P",
    "MOdin10G1S6P_b",
    "MOdin10G1S2PT",
    "MOdin10G1S2P_d",
    "MOdin10G1S12P",
    "MOdin40G2S2P",
    "MOdin40G2S2PB",
    "MOdin100G3S1P",
    "MLoki100G3S1P",
    "MLoki100G3S1P_b",
    "MLoki100G3S1PSE",
    "MLoki100G3S1PB",
    "MLoki100G5S1P",
    "MLoki100G5S2P",
    "MThor100G5S4P",
    "MThor400G7S1P",
    "MThor400G7S1P_b",
    "MThor400G7S1P_c",
    "MThor400G7S1P_d",
    "MFreya800G1S1P_a",
    "MFreya800G1S1P_b",
    "MFreya800G1S1POSFP_a",
    "MFreya800G4S1P_a",
    "MFreya800G4S1P_b",
    "MFreya800G4S1POSFP_a",
    "MOdin1G3S6PT1RJ45",
    "MOdin10G4S2PCombi",
    "MOdin10G4S2PCombi_b",
]

GenericAnyModule = typing.Union[
    GenericL23Module,
    "ModuleL23VE",
    "ModuleChimera",
    "MChi100G5S2P",
    "MChi100G5S2P_b",
    "MChi40G2S2P",
    "ModuleL47",
    "ModuleL47VE",
]

__all__ = (
    "ModuleL23",
    "MOdin1G3S6P",
    "MOdin1G3S6P_b",
    "MOdin1G3S6PE",
    "MOdin1G3S2PT",
    "MOdin5G4S6PCU",
    "MOdin10G5S6PCU",
    "MOdin10G5S6PCU_b",
    "MOdin10G3S6PCU",
    "MOdin10G3S2PCU",
    "MOdin10G1S2P",
    "MOdin10G1S2P_b",
    "MOdin10G1S2P_c",
    "MOdin10G1S2P_d",
    "MOdin10G1S6P",
    "MOdin10G1S6P_b",
    "MOdin10G1S2PT",
    "MOdin10G1S2P_d",
    "MOdin10G1S12P",
    "MOdin40G2S2P",
    "MOdin40G2S2PB",
    "MOdin100G3S1P",
    "MLoki100G3S1P",
    "MLoki100G3S1P_b",
    "MLoki100G3S1PSE",
    "MLoki100G3S1PB",
    "MLoki100G5S1P",
    "MLoki100G5S2P",
    "MThor100G5S4P",
    "MThor400G7S1P",
    "MThor400G7S1P_b",
    "MThor400G7S1P_c",
    "MThor400G7S1P_d",
    "MFreya800G1S1P_a",
    "MFreya800G1S1P_b",
    "MFreya800G1S1POSFP_a",
    "MFreya800G4S1P_a",
    "MFreya800G4S1P_b",
    "MFreya800G4S1POSFP_a",
    "MOdin1G3S6PT1RJ45",
    "MOdin10G4S2PCombi",
    "MOdin10G4S2PCombi_b",
    "ModuleL23VE",
    "ModuleChimera",
    "MChi100G5S2P",
    "MChi100G5S2P_b",
    "MChi40G2S2P",
    "ModuleL47",
    "ModuleL47VE",
    "GenericL23Module",
    "GenericAnyModule",
)



from enum import IntEnum
import asyncio


class ReservedRequestID(IntEnum):
    PUSH_REQUEST = 0
    REQUEST_ID_LIMIT = 0xFFFFFFFF

    @classmethod
    def started(cls) -> int:
        return max(list(cls)[:-1]).value


class RequestId:
    """Aggrigator of request ID."""
    __slots__ = ("__req_id", "__lock",)

    def __init__(self) -> None:
        self.__req_id = ReservedRequestID.started()
        self.__lock = asyncio.Lock()

    async def get_avaliable(self) -> int:
        async with self.__lock:
            if self.__req_id >= ReservedRequestID.REQUEST_ID_LIMIT:
                self.__req_id = ReservedRequestID.started()
            else:
                self.__req_id += 1
            return self.__req_id



from typing import (
    TYPE_CHECKING,
    Union,
)
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
    from xoa_driver.internals.hli_v1.indices.streams.genuine_stream import GenuineStreamIdx
from xoa_driver.internals.core.commands import (
    PT_TOTAL,
    PT_NOTPLD,
    PT_STREAM,
    PT_CLEAR,
    PT_EXTRA,
    PT_TOTALEXT,
    PT_NOTPLDEXT,
    PT_STREAMEXT,
)


class PortTransmissionStatistics:
    """L23 port TX statistics"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.__conn = conn
        self.__module_id = module_id
        self.__port_id = port_id
        self.total = PT_TOTAL(conn, module_id, port_id)
        """All TX statistics on the L23 port.
        
        :type: PT_TOTAL
        """

        self.total_ext = PT_TOTALEXT(conn, module_id, port_id)
        """All TX statistics on the L23 port. (extended)
        
        :type: PT_TOTALEXT
        """

        self.extra = PT_EXTRA(conn, module_id, port_id)
        """Extra TX statistics on the L23 port.
        
        :type: PT_EXTRA
        """

        self.no_tpld = PT_NOTPLD(conn, module_id, port_id)
        """TX statistics of packets without TPLD on the L23 port.
        
        :type: PT_NOTPLD
        """

        self.no_tpld_ext = PT_NOTPLDEXT(conn, module_id, port_id)
        """TX statistics of packets without TPLD on the L23 port. (extended)
        
        :type: PT_NOTPLDEXT
        """

        self.clear = PT_CLEAR(conn, module_id, port_id)
        """Clear TX statistics on the L23 port.
        
        :type: PT_CLEAR
        """

    def obtain_from_stream(self, stream: Union[int, "GenuineStreamIdx"]) -> "PT_STREAM":
        """Obtain statistics of packets of a specific stream on a L23 port.

        :param stream: stream object
        :type stream: Union[int, "GenuineStreamIdx"]
        :return: statistics of packets of a specific stream on a L23 port
        :rtype: PT_STREAM
        """

        stream_idx = stream if isinstance(stream, int) else stream.kind.index_id
        return PT_STREAM(
            self.__conn,
            self.__module_id,
            self.__port_id,
            stream_idx
        )

    def obtain_from_stream_ext(self, stream: Union[int, "GenuineStreamIdx"]) -> "PT_STREAMEXT":
        """Obtain statistics of packets of a specific stream on a L23 port. (extended)

        :param stream: stream object
        :type stream: Union[int, "GenuineStreamIdx"]
        :return: statistics of packets of a specific stream on a L23 port
        :rtype: PT_STREAMEXT
        """

        stream_idx = stream if isinstance(stream, int) else stream.kind.index_id
        return PT_STREAMEXT(
            self.__conn,
            self.__module_id,
            self.__port_id,
            stream_idx
        )



from typing import (
    TYPE_CHECKING,
    Tuple,
)
from dataclasses import astuple
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf

from xoa_driver.internals.core.commands import (
    PE_FCSDROP,
    PE_TPLDMODE,
    PE_COMMENT,
    PE_INDICES,
    PE_LATENCYRANGE,
    PE_CORRUPT,
    PE_MISORDER,
    PE_BANDPOLICER,
    PE_BANDSHAPER,
    PE_DROPTOTAL,
    PE_LATENCYTOTAL,
    PE_DUPTOTAL,
    PE_MISTOTAL,
    PE_CORTOTAL,
    PE_JITTERTOTAL,
    PE_CLEAR,
    PE_FLOWDROPTOTAL,
    PE_FLOWLATENCYTOTAL,
    PE_FLOWDUPTOTAL,
    PE_FLOWMISTOTAL,
    PE_FLOWCORTOTAL,
    PE_FLOWJITTERTOTAL,
    PE_FLOWCLEAR,
)

from .reception_statistics import ReceptionStatistics
from .transmission_statistics import TransmissionStatistics
from .pe_distribution import ImpairmentTypeDistribution
from .filter_definition import (
    shadow,
    working,
)


class CTotalFlow:
    """Total flow statistics."""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, flow_index: int) -> None:
        self.drop_packets = PE_FLOWDROPTOTAL(conn, module_id, port_id, flow_index)
        """Statistics of all packets dropped in a flow.
        
        :type: PE_FLOWDROPTOTAL
        """

        self.latency_packets = PE_FLOWLATENCYTOTAL(conn, module_id, port_id, flow_index)
        """Statistics of all packets delayed in a flow.
        
        :type: PE_FLOWLATENCYTOTAL
        """

        self.duplicated_packets = PE_FLOWDUPTOTAL(conn, module_id, port_id, flow_index)
        """Statistics of all packets duplicate in a flow.
        
        :type: PE_FLOWDUPTOTAL
        """

        self.mis_ordered_packets = PE_FLOWMISTOTAL(conn, module_id, port_id, flow_index)
        """Statistics of all packets misordered in a flow.
        
        :type: PE_FLOWMISTOTAL
        """

        self.corrupted_packets = PE_FLOWCORTOTAL(conn, module_id, port_id, flow_index)
        """Statistics of all packets corrupted in a flow.
        
        :type: PE_FLOWCORTOTAL
        """

        self.jittered_packets = PE_FLOWJITTERTOTAL(conn, module_id, port_id, flow_index)
        """Statistics of all packets jittered in a flow.
        
        :type: PE_FLOWJITTERTOTAL
        """


class StatisticsTotals:
    """Total port statistics."""

    def __init__(self, conn, module_id, port_id) -> None:
        self.drop = PE_DROPTOTAL(conn, module_id, port_id)
        """Statistics of all packets dropped on the port.
        
        :type: PE_DROPTOTAL
        """

        self.latency = PE_LATENCYTOTAL(conn, module_id, port_id)
        """Statistics of all packets delayed on the port.
        
        :type: PE_LATENCYTOTAL
        """

        self.duplicated = PE_DUPTOTAL(conn, module_id, port_id)
        """Statistics of all packets duplicated on the port.
        
        :type: PE_DUPTOTAL
        """

        self.mis_ordered = PE_MISTOTAL(conn, module_id, port_id)
        """Statistics of all packets mirordered on the port.
        
        :type: PE_MISTOTAL
        """

        self.corrupted = PE_CORTOTAL(conn, module_id, port_id)
        """Statistics of all packets corrupted on the port.
        
        :type: PE_CORTOTAL
        """

        self.jittered = PE_JITTERTOTAL(conn, module_id, port_id)
        """Statistics of all packets jittered on the port.
        
        :type: PE_JITTERTOTAL
        """


class CFlowStatistics:
    """Per flow statistics."""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, flow_index: int) -> None:
        self.rx = ReceptionStatistics(conn, module_id, port_id, flow_index)
        """RX statistics."""

        self.tx = TransmissionStatistics(conn, module_id, port_id, flow_index)
        """TX statistics."""

        self.total = CTotalFlow(conn, module_id, port_id, flow_index)
        """Total flow statistics."""

        self.clear = PE_FLOWCLEAR(conn, module_id, port_id, flow_index)
        """Clear the impairment statistics on a flow.
        
        :type: PE_FLOWCLEAR
        """


class CBandwidth:
    """Bandwidth configuration."""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, flow_index: int) -> None:
        self.policer = PE_BANDPOLICER(conn, module_id, port_id, flow_index)
        """Bandwidth policer configuration.
        
        :type: PE_BANDPOLICER
        """

        self.shaper = PE_BANDSHAPER(conn, module_id, port_id, flow_index)
        """Bandwidth shaper configuration.
        
        :type: PE_BANDSHAPER
        """


class CFlow:
    """Flow settings."""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, flow_index: int) -> None:
        self.comment = PE_COMMENT(conn, module_id, port_id, flow_index)
        """Flow description.
        
        :type: PE_COMMENT
        """

        self.latency_range = PE_LATENCYRANGE(conn, module_id, port_id, flow_index)
        """Flow latency range.
        
        :type: PE_LATENCYRANGE
        """

        self.corruption = PE_CORRUPT(conn, module_id, port_id, flow_index)
        """Corruption type.
        
        :type: PE_CORRUPT
        """

        self.misordering = PE_MISORDER(conn, module_id, port_id, flow_index)
        """Misordering depth
        
        :type: PE_MISORDER
        """

        self.bandwidth_control = CBandwidth(conn, module_id, port_id, flow_index)
        """Bandwidth configuration."""

        self.statistics = CFlowStatistics(conn, module_id, port_id, flow_index)
        """Flow statistics."""

        self.impairment_distribution = ImpairmentTypeDistribution(conn, module_id, port_id, flow_index)
        """Impairment type's distribution."""

        self.shadow_filter = shadow.FilterDefinitionShadow(conn, module_id, port_id, flow_index)
        """Shadow copy."""

        self.working_filter = working.FilterDefinitionWorking(conn, module_id, port_id, flow_index)
        """Working copy."""


class ChimeraPE:
    """Port Emulation"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self._conn = conn
        self.module_id = module_id
        self.port_id = port_id
        self.drop_fcs_errors = PE_FCSDROP(conn, module_id, port_id)
        """Action on FCS errors.
        
        :type: PE_FCSDROP
        """

        self.clear = PE_CLEAR(conn, module_id, port_id)
        """Clear impairment statistics.
        
        :type: PE_CLEAR
        """

        self.statistics = StatisticsTotals(conn, module_id, port_id)
        """Total port statistics."""

        self.tpld_mode = PE_TPLDMODE(conn, module_id, port_id)
        """TPLD mode.
        
        :type: PE_TPLDMODE
        """

        self.flows: Tuple["CFlow", ...] = tuple()

    def __await__(self):
        return self._setup().__await__()

    async def _setup(self) -> None:
        indices = astuple(await PE_INDICES(self._conn, self.module_id, self.port_id).get())
        self.flows = tuple(
            CFlow(self._conn, self.module_id, self.port_id, idx)
            for idx in indices
        )



from typing import TYPE_CHECKING
from xoa_driver.internals.core.commands import P_AUTONEGSELECTION
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf

from .bases.port_l23_genuine import BasePortL23Genuine
from . import family_f


class POdin1G4S4PCombi(BasePortL23Genuine):
    """L23 ports that supports MDI/MDIX and """
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        super().__init__(conn, module_id, port_id)

        self.autoneg_selection = P_AUTONEGSELECTION(conn, module_id, port_id)
        """L23 port's auto-negotiation selection.
        
        :type: P_AUTONEGSELECTION`
        """


class POdin1G4S4PCombi_b(POdin1G4S4PCombi):
    """L23 port on Odin-1G-4S-2P-Combi_b module.
    """
    ...


class POdin10G4S2PCombi(family_f.FamilyF):
    """L23 port on Odin-10G-4S-2P-Combi module.
    """
    ...


class POdin10G4S2PCombi_b(family_f.FamilyF):
    """L23 port on Odin-10G-4S-2P-Combi_b module.
    """
    ...



from asyncio.futures import Future
from typing import (
    Dict,
    TYPE_CHECKING,
    Tuple,
    Optional,
)
if TYPE_CHECKING:
    from .. import protocol


class CommandsManager:
    __slots__ = ("__cmd_handler", "__received_futures")

    def __init__(self) -> None:
        self.__cmd_handler: Dict[int, int] = {}
        self.__received_futures: Dict[Tuple[int, str], Future] = {}

    def append(self, request: "protocol.Request") -> "Future":
        self.__cmd_handler[request.header.request_identifier] = request.header.cmd_code
        self.__received_futures[
            (
                request.header.request_identifier,
                request.class_name
            )
        ] = fut = Future()
        return fut

    def get_command_handler_id(self, req_id) -> Optional[int]:
        return self.__cmd_handler.pop(req_id, None)

    def get_result_future(self, response) -> Optional["Future"]:
        return self.__received_futures.pop(
            (
                response.header.request_identifier,
                response.class_name
            ),
            None
        )



#: Enumeration Types

from enum import IntEnum


# region L23 enums
class ReservedStatus(IntEnum):
    """Test resource reservation status"""

    RELEASED = 0
    """Test resource is released."""
    RESERVED_BY_YOU = 1
    """Test resource is reserved by you."""
    RESERVED_BY_OTHER = 2
    """Test resource is reserved by another user."""


class ReservedAction(IntEnum):
    """Reservation actions."""

    RELEASE = 0
    """Release the test resource."""
    RESERVE = 1
    """Reserve the test resource."""
    RELINQUISH = 2
    """Force release other's ownership of the test resource."""


class ChassisShutdownAction(IntEnum):
    """Chassis shutdown actions."""

    RESTART = 1
    """Restart after shutdown."""
    POWER_OFF = 2
    """Keep power-off after shutdown."""


class OnOff(IntEnum):
    """On or Off."""

    OFF = 0
    """Off"""
    ON = 1
    """On"""


class RESTControlAction(IntEnum):
    """Controls of Valkyrie REST server."""

    START = 0
    """Start the REST server on the chassis."""
    STOP = 1
    """Stop the REST server on the chassis."""
    RESTART = 2
    """Restart the REST server on the chassis."""


class ServiceStatus(IntEnum):
    """Service status."""

    SERVICE_OFF = 0
    """Service is off."""
    SERVICE_ON = 1
    """Service is on."""


class ChassisSessionType(IntEnum):
    """The type of the session between the client and the chassis."""

    MANAGER = 1
    """Connection on port 22606 running binary command (used by ValkyrieManager & VulcanManager)."""
    SCRIPT = 2
    """Connection on port 22611, running CLI command."""


class TimingSource(IntEnum):
    """Module's time sync mode."""

    CHASSIS = 0
    """Module syncs to chassis's clock."""
    EXTERNAL = 1
    """Module syncs to an external clock."""
    MODULE = 2
    """Module syncs to its own clock."""


class MediaCFPState(IntEnum):
    """Modules' CFP state"""

    NOT_CFP = 0
    """This is not a CFP-based test module."""
    NOT_PRESENT = 1
    """No transceiver, the CFP cage is empty"""
    NOT_FLEXIBLE = 2
    """Transceiver present, supporting a fixed speed and port-count"""
    FLEXIBLE = 3
    """Transceiver present, supporting flexible speed and port-count"""


class MediaCFPType(IntEnum):
    """Module's Media CFP Type. What kind of transceiver, may have multiple ports"""

    CFP_UNKNOWN = 0
    """CFP unknown"""
    CFP_INVALID = 1
    """CFP 100G not supported"""
    CFP_1X40_SR4 = 2
    """CFP 40G SR4 850 nm"""
    CFP_2X40_SR4 = 3
    """CFP 40G SR4 850 nm"""
    CFP_1X40_LR4 = 4
    """CFP 40G LR4 1310 nm"""
    CFP_1X100_LR4 = 5
    """CFP 100G LR4 1310 nm"""
    CFP_1X100_SR10 = 6
    """CFP 100G SR10 850 nm / CFP 40G SR4 850 nm"""
    CFP_4X10_SFPP = 7
    """"""
    QSFP_EMPTY = 8
    """QSFP empty"""
    QSFP_UNKNOWN = 9
    """QSFP unknown"""
    QSFP_SR = 10
    """QSFP SR 850 nm"""
    QSFP_LR = 11
    """QSFP LR 1550 nm"""
    QSFP_ER = 12
    """QSFP ER 1550 nm"""
    CXP_UNKNOWN = 13
    """CXP unknown SR10"""
    CXP_SR10 = 14
    """CXP 100G SR10 850 nm / CXP 40G SR4 850 nm"""
    CXP_AOC = 15
    """CXP 100G SR10 AOC / CXP 40G SR4 AOC"""
    CFP4_UNKNOWN = 16
    """CFP4 unknown"""
    CFP4_1X100_LR4 = 17
    """CFP4 100G LR4"""
    CFP4_1X100_ER4 = 18
    """CFP4 100G ER4"""
    CFP4_1X100_SR4 = 19
    """CFP4 100G SR4"""
    QSFP28_CR4 = 20
    """QSFP28 CR4"""
    QSFP28_SR4 = 21
    """QSFP28 SR4"""
    QSFP28_LR4 = 22
    """QSFP28 LR4"""
    QSFP28_ER4 = 23
    """QSFP28 ER4"""
    QSFP28_AOC = 24
    """QSFP28 AOC"""
    QSFP28_ACC = 25
    """QSFP28 ACC"""
    QSFP28_UNKNOWN = 26
    """QSFP28 unknown"""
    QSFP28P_SR4 = 27
    """QSFP+ SR4 850 nm"""
    QSFP28P_LR4 = 28
    """QSFP+ LR4 1310 nm"""
    QSFP28P_ER4 = 29
    """QSFP+ ER4 1500 nm"""
    QSFP28P_CR4 = 30
    """QSFP+ CR4 Copper"""
    QSFP28P_ISM4 = 31
    """QSFP+ iSM4 1310 nm"""
    QSFP28P_UNKNOWN = 32
    """QSFP+ unknown"""
    TCVR_UNKNOWN = 33
    """Transceiver unknown"""
    QSFP28_CWDM4 = 34
    """QSFP28 CWDM4"""
    QSFP28_CWDM4_FEC = 35
    """QSFP28 CWDM4 FEC"""
    QSFP28_PSM4 = 36
    """QSFP28 PSM4"""


class SMAInputFunction(IntEnum):
    """SMA input function"""

    NOT_USED = 0
    """SMA input not used"""
    TX2MHZ = 1
    """TX Clock Ref. 2.048 MHz"""
    TX10MHZ = 2
    """TX Clock Ref. 10.0 MHz"""


class SMAOutputFunction(IntEnum):
    """SMA output function"""

    DISABLED = 0
    """Disabled"""
    PASSTHROUGH = 1
    """Pass-through"""
    P0SOF = 2
    """Port 0 Start-of-Frame Pulse"""
    P1SOF = 3
    """Port 1 Start-of-Frame Pulse"""
    REF2MHZ = 4
    """TX Clock (nom. 2.048 MHz)"""
    REF10MHZ = 5
    """TX Clock (nom. 10.0 MHz)"""
    REF125MHZ = 6
    """TX Clock (nom. 125 MHz)"""
    REF156MHZ = 7
    """TX Clock (nom. 156.25 MHz)"""
    P0RXCLK = 8
    """Port 0 RX Clock (nom. 10.0 MHz)"""
    P1RXCLK = 9
    """Port 1 RX Clock (nom. 10.0 MHz)"""
    P0RXCLK2MHZ = 10
    """Port 0 RX Clock (nom. 2.048 MHz)"""
    P1RXCLK2MHZ = 11
    """Port 1 RX Clock (nom. 2.048 MHz)"""
    TS_PPS = 12
    """Timing Source (Pulse-Per-Second)"""


class SMAStatus(IntEnum):
    """SMA status"""

    OK = 0
    """Status OK"""
    NO_VALID_SIGNAL = 1
    """No valid signal"""


class HasDemo(IntEnum):
    """Demo or not"""

    NON_DEMO = 0
    """Not a demo unit"""
    DEMO = 1
    """Demo unit"""


class IsValid(IntEnum):
    """Whether it is valid"""

    INVALID = 0
    """Invalid"""
    VALID = 1
    """Valid"""


class IsPermanent(IntEnum):
    """Whether it is permanent"""

    NON_PERMANENT = 0
    """Not permanent"""
    PERMANENT = 1
    """Permanent"""


class YesNo(IntEnum):
    NO = 0
    """No"""
    YES = 1
    """Yes"""


class UpdateState(IntEnum):
    """License update state"""

    NONE = 0
    """None"""
    UPDATING = 1
    """Updating"""
    UPDATE_SUCCESS = 2
    """Update successful"""
    UPDATE_FAIL = 3
    """Update failed"""


class IsOnline(IntEnum):
    """Chassis online or not"""

    OFFLINE = 0
    """Chassis offline"""
    ONLINE = 1
    """Chassis online"""


class PortSpeedMode(IntEnum):
    """Port speed mode"""

    AUTO = 0
    """Auto negotiate with peer"""
    F10M = 1
    """Forced to 10 Mbit/s"""
    F100M = 2
    """Forced to 100 Mbit/s"""
    F1G = 3
    """Forced to 1 Gbit/s"""
    F10G = 4
    """Forced to 10 Gbit/s"""
    F40G = 5
    """Forced to 40 Gbit/s"""
    F100G = 6
    """Forced to 100 Gbit/s"""
    F10MHDX = 7
    """Forced to 10 Mbit/s half-duplex"""
    F100MHDX = 8
    """Forced to 100 Mbit/s half-duplex"""
    F10M100M = 9
    """Auto negotiate to 10/100 Mbit/s"""
    F100M1G = 10
    """Auto negotiate to 100 Mbit/s or 1 Gbit/s"""
    F100M1G10G = 11
    """Auto negotiate to 100 Mbit/s or 1 Gbit/s or 10 Gbit/s"""
    F2500M = 12
    """Forced to 2500 Mbit/s"""
    F5G = 13
    """Forced to 5 Gbit/s"""
    F100M1G2500M = 14
    """Auto negotiate to 100 Mbit/s or 1 Gbit/s or 2500 Mbit/s"""
    F25G = 15
    """Forced to 25 Gbit/s"""
    F50G = 16
    """Forced to 50 Gbit/s"""
    F200G = 17
    """Forced to 200 Gbit/s"""
    F400G = 18
    """Forced to 400 Gbit/s"""
    F800G = 19
    """Forced to 800 Gbit/s"""
    F1600G = 20
    """Forced to 1600 Gbit/s"""
    UNKNOWN = 255
    """Speed mode unknown"""


class SyncStatus(IntEnum):
    """Port sync status"""

    NO_SYNC = 0
    """Not sync"""
    IN_SYNC = 1
    """In sync with peer"""


class LoopbackMode(IntEnum):
    """Port loopback mode"""

    NONE = 0
    """Off"""
    L1RX2TX = 1
    """L1 RX-to-TX"""
    L2RX2TX = 2
    """L2 RX-to-TX"""
    L3RX2TX = 3
    """L3 RX-to-TX"""
    TXON2RX = 4
    """TX(on)-to-RX"""
    TXOFF2RX = 5
    """TX(off)-to-RX"""
    PORT2PORT = 6
    """Port-to-Port"""


class TrafficOnOff(IntEnum):
    """Traffic status"""

    OFF = 0
    """Traffic off"""
    ON = 1
    """Traffic on"""
    PREPARING = 2
    """Traffic preparing"""


class StartOrStop(IntEnum):
    STOP = 0
    """Stop the action"""
    START = 1
    """Start the action"""


class LatencyMode(IntEnum):
    """Latency measurement mode"""

    LAST2LAST = 0
    """Last-to-Last"""
    FIRST2LAST = 1
    """First-to-Last"""
    LAST2FIRST = 2
    """Last-to-First"""
    FIRST2FIRST = 3
    """First-to-First"""


class SourceType(IntEnum):
    TX_IFG = 0
    """TX IFG"""
    TX_LEN = 1
    """TX Length"""
    RX_IFG = 2
    """RX IFG"""
    RX_LEN = 3
    """RX Length"""
    RX_LATENCY = 4
    """RX Latency"""
    RX_JITTER = 5
    """RX Jitter"""


class PacketDetailSelection(IntEnum):
    ALL = 0
    """All"""
    TPLD = 1
    """Based on Test Payload ID"""
    FILTER = 2
    """Based on filter"""


class OnOffWithSuppress(IntEnum):
    """Stream status"""

    OFF = 0
    """Off"""
    ON = 1
    """On"""
    SUPPRESS = 2
    """Suppressed"""


class ProtocolOption(IntEnum):
    """Protocol header options"""

    ETHERNET = 1
    """Ethernet II"""
    VLAN = 2
    """VLAN"""
    ARP = 3
    """Address Resolution Protocol"""
    IP = 4
    """IPv4"""
    IPV6 = 5
    """IPv6"""
    UDP = 6
    """User Datagram Protocol (w/o checksum)"""
    TCP = 7
    """Transmission Control Protocol (w/o checksum)"""
    LLC = 8
    """Logic Link Control"""
    SNAP = 9
    """Subnetwork Access Protocol"""
    GTP = 10
    """GPRS Tunnelling Protocol"""
    ICMP = 11
    """Internet Control Message Protocol"""
    RTP = 12
    """Real-time Transport Protocol"""
    RTCP = 13
    """RTP Control Protocol"""
    STP = 14
    """Spanning Tree Protocol"""
    SCTP = 15
    """Stream Control Transmission Protocol"""
    MACCTRL = 16
    """MAC Control"""
    MPLS = 17
    """Multiprotocol Label Switching"""
    PBBTAG = 18
    """Provider Backbone Bridge tag"""
    FCOE = 19
    """Fibre Channel over Ethernet"""
    FC = 20
    """Fibre Channel"""
    FCOETAIL = 21
    """Fibre Channel over Ethernet (tail)"""
    IGMPV3L0 = 22
    """IGMPv3 Membership Query L=0"""
    IGMPV3L1 = 23
    """IGMPv3 Membership Query L=1"""
    UDPCHECK = 24
    """User Datagram Protocol (w/ checksum)"""
    IGMPV2 = 25
    """Internet Group Management Protocol v2"""
    MPLS_TP_OAM = 26
    """MPLS-TP, OAM Header"""
    GRE_NOCHECK = 27
    """Generic Routing Encapsulation (w/o checksum)"""
    GRE_CHECK = 28
    """Generic Routing Encapsulation (w/ checksum)"""
    TCPCHECK = 29
    """Transmission Control Protocol (w/ checksum)"""
    GTPV1L0 = 30
    """GTPv1 (no options), GPRS Tunneling Protocol v1"""
    GTPV1L1 = 31
    """GTPv1 (w/ options), GPRS Tunneling Protocol v1"""
    GTPV2L0 = 32
    """GTPv2 (no options), GPRS Tunneling Protocol v2"""
    GTPV2L1 = 33
    """GTPv2 (w/ options), GPRS Tunneling Protocol v2"""
    IGMPV1 = 34
    """Internet Group Management Protocol v1"""
    PWETHCTRL = 35
    """PW Ethernet Control Word"""
    VXLAN = 36
    """Virtual eXtensible LAN"""
    ETHERNET_8023 = 37
    """Ethernet 802.3"""
    NVGRE = 38
    """Generic Routing Encapsulation (Network Virtualization)"""
    DHCPV4 = 39
    """Dynamic Host Configuration Protocol (IPv4)"""
    GENEVE = 40
    """Generic Network Virtualization Encapsulation"""
    XENA_TPLD = 41
    XENA_TPLD_PI = 42
    XENA_MICROTPLD = 43
    ETHERNET_FCS = 44
    MACCTRLPFC = 45
    """MAC Control for PFC"""
    ECPRI = 46
    """Enhanced Common Public Radio Interface"""
    ROE = 47
    """Radio over Ethernet"""
    ETHERTYPE = 48
    """EtherType"""

    # Generate RAW form 1...64 bytes
    _ignore_ = "ProtocolOption i"
    ProtocolOption = vars()
    for i in range(1, 65):
        ProtocolOption["RAW_%d" % i] = 256 - i  # type: ignore


class ModifierAction(IntEnum):
    """Modifier action mode"""

    INC = 0
    """Incrementing"""
    DEC = 1
    """Decrementing"""
    RANDOM = 2
    """Random"""


class LengthType(IntEnum):
    """Packet length type"""

    FIXED = 0
    """Fixed"""
    INCREMENTING = 1
    """Incrementing"""
    BUTTERFLY = 2
    """Butterfly"""
    RANDOM = 3
    """Random"""
    MIX = 4
    """Mix"""


class PayloadType(IntEnum):
    """Packet payload type"""

    PATTERN = 0
    """Pattern"""

    INC8 = INCREMENTING = 1
    """
    .. versionchanged:: 1.1

    Incrementing with 0xFF (8-bit mode)

    """

    PRBS = 2
    """PRBS"""

    RANDOM = 3
    """Random"""

    DEC8 = DECREMENTING = 4
    """
    .. versionchanged:: 1.1

    Decrementing with 0xFF (8-bit mode)
    """

    INC16 = 5
    """
    .. versionadded:: 1.1

    Incrementing with 0xFFFF (16-bit mode)
    """

    DEC16 = 6
    """
    .. versionadded:: 1.1

    Decrementing with 0xFFFF (16-bit mode)
    """


class MDIXMode(IntEnum):
    """MDIX mode"""

    AUTO = 0
    """Auto"""
    MDI = 1
    """MDI"""
    MDIX = 2
    """MDIX"""


class LengthCheckType(IntEnum):
    AT_MOST = 0
    """At Most"""
    AT_LEAST = 1
    """At Least"""


class StartTrigger(IntEnum):
    """Capture start trigger"""

    ON = 0
    """From Traffic On"""
    FCSERR = 1
    """From FCS Error"""
    FILTER = 2
    """From Filter"""
    PLDERR = 3
    """From Payload Error"""


class StopTrigger(IntEnum):
    """Capture stop trigger"""

    FULL = 0
    """Until Capture Buffer Full"""
    FCSERR = 1
    """Until Receiving FCS Error"""
    FILTER = 2
    """Until Filter Matched"""
    PLDERR = 3
    """Until Receiving Payload Error"""
    USERSTOP = 4
    """Until User Stops"""


class PacketType(IntEnum):
    """Type of Packet to Keep in Capture Buffer"""

    ALL = 0
    """All Packets"""
    FCSERR = 1
    """With FCS Errors"""
    NOTPLD = 2
    """Without Test Payload"""
    TPLD = 3
    """With Test Payload"""
    FILTER = 4
    """Filtered Packets"""
    PLDERR = 5
    """With Payload Errors"""


class InjectErrorType(IntEnum):
    """Lane Injection Error Type"""

    HEADERERROR = 1
    """Header Error"""
    ALIGNERROR = 2
    """Alignment Error"""
    BIP8ERROR = 3
    """BIP8 Error"""


class HeaderLockStatus(IntEnum):
    """Physical Lane Header Lock Status"""

    HEADEROFF = 0
    """Header Lock Off"""
    HEADERON = 1
    """Header Lock On"""
    HEADEROFFUNSTABLE = 2
    """Header Lock Off and Unstable"""
    HEADERONUNSTABLE = 3
    """Header Lock On but Unstable"""


class AlignLockStatus(IntEnum):
    """Physical Lane Alignment Lock Status"""

    ALIGNOFF = 0
    """Alignment Lock Off"""
    ALIGNON = 1
    """Alignment Lock On"""
    ALIGNOFFUNSTABLE = 2
    """Alignment Lock Off and Unstable"""
    ALIGNONUNSTABLE = 3
    """Alignment Lock On but Unstable"""


class PRBSLockStatus(IntEnum):
    """Physical Lane PRBS Lock Status"""

    PRBSOFF = 0
    """PRBS Lock Off"""
    PRBSON = 1
    """PRBS Lock On"""
    PRBSOFFUNSTABLE = 2
    """PRBS Lock Off and Unstable"""
    PRBSONUNSTABLE = 3
    """PRBS Lock On but Unstable"""


class MulticastOperation(IntEnum):
    """IGMPv2 Request Type"""

    OFF = 0
    """Off"""
    ON = 1
    """On"""
    JOIN = 2
    """Join"""
    LEAVE = 3
    """Leave"""


class MulticastExtOperation(IntEnum):
    """IGMPv2/v3Request Type"""

    OFF = 0
    """Off"""
    ON = 1
    """On"""
    JOIN = 2
    """Join"""
    LEAVE = 3
    """Leave"""
    INCLUDE = 4
    """Include"""
    EXCLUDE = 5
    """EXclude"""
    LEAVE_TO_ALL = 6
    """Leave To All"""
    GENERAL_QUERY = 7
    """General Query"""
    GROUP_QUERY = 8
    """Group Query"""


class IGMPVersion(IntEnum):
    """IGMP Version"""

    IGMPV2 = 0
    """IGMP Version 2"""
    IGMPV3 = 1
    """IGMP Version 3"""


class TXMode(IntEnum):
    """Port TX Mode"""

    NORMAL = 0
    """Normal"""
    STRICTUNIFORM = 1
    """Strict Uniform"""
    SEQUENTIAL = 2
    """Sequential"""
    BURST = 3
    """Burst"""


class PayloadMode(IntEnum):
    """Payload Mode"""

    NORMAL = 0
    """Normal"""
    EXTPL = 1
    """Extend Payload"""
    CDF = 2
    """Custom Data Field"""


class BRRMode(IntEnum):
    """BRR Mode"""

    SLAVE = 0
    """Slave Mode"""
    MASTER = 1
    """Master Mode"""


class TXClockSource(IntEnum):
    """TX Clock Source"""

    MODULELOCALCLOCK = 0
    """Module Local Clock"""
    SMAINPUT = 1
    """SMA Input"""
    P0RXCLK = 2
    """Port 0 RX Clock"""
    P1RXCLK = 3
    """Port 1 RX Clock"""
    P2RXCLK = 4
    """Port 2 RX Clock"""
    P3RXCLK = 5
    """Port 3 RX Clock"""
    P4RXCLK = 6
    """Port 4 RX Clock"""
    P5RXCLK = 7
    """Port 5 RX Clock"""
    P6RXCLK = 8
    """Port 6 RX Clock"""
    P7RXCLK = 9
    """Port 7 RX Clock"""


class TXClockStatus(IntEnum):
    """TX Clock Status"""

    OK = 0
    """Clock OK"""
    NOVALIDTXCLK = 1
    """No Valid TX Clock"""


class LoopBandwidth(IntEnum):
    """Loop Bandwidth"""

    BW103HZ = 1
    """Loop Bandwidth = 103 Hz"""
    BW207HZ = 2
    """Loop Bandwidth = 207 Hz"""
    BW416HZ = 3
    """Loop Bandwidth = 416 Hz"""
    BW1683HZ = 4
    """Loop Bandwidth = 1683 Hz"""
    BW7019HZ = 5
    """Loop Bandwidth = 7019 Hz"""


class MediaConfigurationType(IntEnum):
    """Module Media Configuration Type"""

    CFP4 = 0
    """CFP4"""

    QSFP28 = QSFP28_NRZ = 1
    """
    .. versionchanged:: 1.1

    QSFP28 (NRZ)
    """

    CXP = 2
    """CXP"""

    SFP28 = 3
    """SFP28"""

    QSFP56 = QSFP56_PAM4 = 4
    """
    .. versionchanged:: 1.1

    QSFP56 (PAM4)
    """

    QSFPDD = QSFPDD_PAM4 = 5
    """
    .. versionchanged:: 1.1

    QSFP-DD (PAM4)
    """

    SFP56 = 6
    """SFP56"""

    SFP_DD = SFPDD = 7
    """
    .. versionchanged:: 1.1

    SFP-DD
    """

    SFP112 = 8
    """SFP112"""

    QSFP_DD_NRZ = QSFPDD_NRZ = 9
    """
    .. versionchanged:: 1.1

    QSFP-DD (NRZ)
    """

    QSFP28_PAM4 = 10
    """QSFP28 (PAM4)"""

    CFP = 99
    """CFP"""

    BASE_T1 = 100
    """BASE-T1"""

    BASE_T1S = 101
    """BASE-T1S"""

    QSFPDD800 = 110
    """QSFP-DD800"""

    QSFP112 = 111
    """QSFP112"""

    OSFP800 = 112
    """OSFP800"""

    UNKNOWN = 255


class TXHState(IntEnum):
    """Recent Change in EEE State on TX"""

    TXH_NA = 0
    """Leaving or Going Into"""
    TXH_X = 1
    """No Activity"""


class RXHState(IntEnum):
    """Recent Change in EEE State on RX"""

    RXH_NA = 0
    """Leaving or Going Into"""
    RXH_X = 1
    """No Activity"""


class TXCState(IntEnum):
    """TX EEE State"""

    TXC_ACTIVE = 0
    """Active"""
    TXC_LPI = 1
    """Low Power"""


class RXCState(IntEnum):
    """RX EEE State"""

    RXC_ACTIVE = 0
    """Active"""
    RXC_LPI = 1
    """Low Power"""


class LinkState(IntEnum):
    """Low Power Mode Link State"""

    LINK_DOWN = 0
    """Link Down"""
    LINK_UP = 1
    """Link Up"""


class FaultSignaling(IntEnum):
    """Fault Signaling Behavior"""

    NORMAL = 0
    """Normal"""
    FORCE_LOCAL = 1
    """Forced to Local"""
    FORCE_REMOTE = 2
    """Forced to Remote"""
    DISABLED = 3
    """Disabled"""


class LocalFaultStatus(IntEnum):
    """Local Fault Status"""

    OK = 0
    """OK"""
    LOCAL_FAULT = 1
    """Local Fault"""


class RemoteFaultStatus(IntEnum):
    """Remote Fault Status"""

    OK = 0
    """OK"""
    REMOTE_FAULT = 1
    """Remote Fault"""


class TPLDMode(IntEnum):
    """Test Payload Mode"""

    NORMAL = 0
    """Normal"""
    MICRO = 1
    """Micro"""


class SerdesStatus(IntEnum):
    """Serdes Status"""

    STOPPED = 0
    """Stopped"""
    STARTED = 1
    """Started"""
    INITIALIZING = 2
    """Initializing"""
    PLOTTING = 3
    """Plotting"""


class FECMode(IntEnum):
    """FEX Mode"""

    OFF = 0
    """Off"""
    ON = 1
    """On"""
    RS_FEC = 2
    """RS FEC"""
    FC_FEC = 3
    """Firecode FEC"""
    RS_FEC_KR = 4
    """RS FEC KR"""
    RS_FEC_KP = 5
    """RS FEC KP"""


class PRBSInsertedType(IntEnum):
    """PRBS Type"""

    CAUI_VIRTUAL = 0
    """CAUI Virtual"""
    PHY_LINE = 1
    """PHY Line"""
    PHY_HOST = 2
    """PHY Host"""
    TCVR = 3
    """Transceiver"""


class PRBSPolynomial(IntEnum):
    """PRBS Polynomial"""

    PRBS7 = 0
    """PRBS-7"""
    PRBS9 = 1
    """PRBS-9"""
    PRBS11 = 2
    """PRBS-11"""
    PRBS15 = 3
    """PRBS-15"""
    PRBS23 = 4
    """PRBS-23"""
    PRBS31 = 5
    """PRBS-31"""
    PRBS58 = 6
    """PRBS-58"""
    PRBS49 = 7
    """PRBS-49"""
    PRBS10 = 8
    """PRBS-10"""
    PRBS20 = 9
    """PRBS-20"""
    PRBS13 = 10
    """PRBS-13"""


class PRBSInvertState(IntEnum):
    """PRBS Invert State"""

    NON_INVERTED = 0
    """Non-inverted"""
    INVERTED = 1
    """Inverted"""


class PRBSStatisticsMode(IntEnum):
    """PRBS Statistics Mode"""

    ACCUMULATIVE = 0
    """Accumulative"""
    PERSECOND = 1
    """Per Second"""


class PauseMode(IntEnum):
    """Pause Mode"""

    NO_PAUSE = 0
    """No Pause"""
    SYM_PAUSE = 1
    """Symmetric Pause"""
    ASYM_PAUSE = 2
    """Asymmetric Pause"""


class MulticastHeaderFormat(IntEnum):
    """Additional Header to IGMPv2/v3 Packets"""

    NOHDR = 0
    """No Header"""
    VLAN = 1
    """VLAN"""


class PFCMode(IntEnum):
    """Priority Flow Control (PFC) Mode"""

    VLAN_PCP = 128
    """VLAN PCP"""


class PRBSOnOff(IntEnum):
    """PRBS Status"""

    PRBSOFF = 0
    """PRBS Off"""
    PRBSON = 1
    """PRBS On"""


class ErrorOnOff(IntEnum):
    """PRBS Error Injection Status"""

    ERRORSOFF = 0
    """PRBS Error Injection Off"""
    ERRORSON = 1
    """PRBS Error Injection On"""


class PRBSPattern(IntEnum):
    """PRBS Pattern"""

    PRBS7 = 0
    """PRBS-7"""
    PRBS9 = 1
    """PRBS-9"""
    PRBS11 = 2
    """PRBS-11"""
    PRBS15 = 3
    """PRBS-15"""
    PRBS23 = 4
    """PRBS-23"""
    PRBS31 = 5
    """PRBS-31"""


class PHYSignalStatus(IntEnum):
    """PHY Signal Status"""

    NO_SIGNAL = 0
    """No Signal"""
    NO_CDRLOCK = 2
    """No CDR Lock"""
    LOCKED = 3
    """Locked"""


class OnOffDefault(IntEnum):
    """On Off Default Status"""

    OFF = 0
    """Off"""
    ON = 1
    """On"""
    DEFAULT = 2
    """Default"""


class TimeKeeperLicenseFileState(IntEnum):
    """TimeKeeper License File State"""

    NA = 0
    """Not Available"""
    INV = 1
    """Invalid"""
    VALID = 2
    """Valid"""


class TimeKeeperLicenseType(IntEnum):
    """TimeKeeper License Type"""

    UNDEF = 0
    """Undefined"""
    CLIENT = 1
    """TimeKeeper Client"""
    SERVER = 2
    """TimeKeeper Server"""


class TimeKeeperLicenseError(IntEnum):
    """TimeKeeper License Error"""

    NO_LICENSE_ERROR = 0
    """No License Error"""
    INVALID_SERIALNO = 1
    """Invalid Serial Number"""
    INVALID_CHASSISTYPE = 2
    """Invalid Chassis Type"""


class SystemUpdateStatus(IntEnum):
    """System Update Status"""

    OK = 0
    """Update OK"""
    FAILED_SCRIPT = 1
    """Script Failed"""
    FAILED_PREP = 10
    """Preparation Failed"""
    FAILED_FILENAME = 11
    """Filename Failed"""
    FAILED_DECRYPT = 12
    """Decryption Failed"""
    FAILED_UNPACK = 13
    """Unpacking Failed"""
    FAILED_VERIFY = 14
    """Verification Failed"""
    FAILED_MOVE = 15
    """Moving Failed"""


class TimeKeeperServiceStatus(IntEnum):
    """TimeKeeper Service Status"""

    STOPPED = 0
    """Service Stopped"""
    STARTED = 1
    """Service Started"""
    NA = 2
    """Not Available"""


class TimeKeeperServiceAction(IntEnum):
    """TimeKeeper Service Action"""

    STOP = 0
    """Stop"""
    START = 1
    """Start"""
    RESTART = 2
    """Restart"""


class CustomDefaultCommand(IntEnum):
    """Custom Default Command"""

    SET = 0
    """Set Custom Default"""
    CLEAR = 1
    """Clear Custom Default"""


class CustomDefaultScope(IntEnum):
    """Custom Default Scope"""

    ALL = 0
    """All"""
    INSTANCE = 1
    """Instance"""


class TrafficError(IntEnum):
    """Traffic Error"""

    NOT_PREPARED = 0
    """Not Prepared"""
    RATE_LENGTH_ERROR = 1
    """Rate Length Error"""
    PREPARED_OK = 2
    """Prepared OK"""


class TrafficEngine(IntEnum):
    """Traffic Engine"""

    TGA = 1
    """Normal TGA"""
    MICRO_TGA = 2
    """Micro TGA"""


class LinkTrainFrameLock(IntEnum):
    """L1 Link Training Frame Lock Status"""

    LOST = 0
    """No Frame Lock detected"""

    LOCKED = 1
    """Frame Lock detected"""



class PPMSweepMode(IntEnum):
    """Module clock PPM Sweep Modes"""

    OFF = 0
    """Off"""
    TRIANGLE = 1
    """Triangle sweeping"""


class PPMSweepStatus(IntEnum):
    """Module clock PPM Sweep Status"""

    OFF = 0
    """Off"""
    SWEEPING = 1
    """The module is sweeping"""


class ReconciliationSublayerSupport(IntEnum):
    """Reconciliation Sublayer Support"""

    NO_SUPPORT = 0
    """Not Supported"""
    FAULT_SIGNALING = 1
    """Supported, which means P_FAULTSTATUS and P_FAULTSIGNALLING are supported by the port."""


# endregion

# region L47 enums
class Role(IntEnum):
    """L47 port role"""

    CLIENT = 0
    """Client"""
    SERVER = 1
    """Server"""


class Timescale(IntEnum):
    """Time scale"""

    MSECS = 0
    """Milliseconds"""
    SECONDS = 1
    """Seconds"""
    MINUTES = 2
    """Minutes"""
    HOURS = 3
    """Hours"""


class MSSType(IntEnum):
    """TCP Maximum Segment Type"""

    FIXED = 0
    """Fixed"""
    INCREMENT = 1
    """Incrementing"""
    RANDOM = 2
    """Pseudorandom"""


class RTOType(IntEnum):
    """TCP RTO Type"""

    STATIC = 0
    """Static RTO"""
    DYNAMIC = 1
    """Dynamic RTO"""


class CongestionType(IntEnum):
    """TCP Congestion Control Algorithm"""

    NONE = 0
    """No TCP Congestion Control"""
    RENO = 1
    """RENO"""
    NEW_RENO = 2
    """New RENO"""


class IsEnabled(IntEnum):
    DISABLE = 0
    ENABLE = 1


class AlgorithmMethod(IntEnum):
    """Algorithm to calculate the TCP initial congestion window (ICWND)"""

    RFC5681 = 0
    """RFC 5681"""
    RFC2581 = 1
    """RFC 2581"""
    FIXED_FACTOR = 2
    """Fixed"""


class AutoOrManual(IntEnum):
    AUTOMATIC = 0
    MANUAL = 1


class EmbedIP(IntEnum):
    """Should Embed IP in MAC"""

    DONT_EMBED_IP = 0
    """Do not embed"""
    EMBED_IP = 1
    """Embed IP in MAC"""


class ApplicationLayerBehavior(IntEnum):
    """L47 Test Application Type"""

    NONE = 0
    """TCP connections are created according to the client and server ranges, and ramped up/down as specified in the load profile. But no payload is transmitted."""
    RAW = 1
    """Differs from ``NONE`` in that it transmits payload when the TCP connection is established."""
    REPLAY = 2
    """PCAP replay."""


class TrafficScenario(IntEnum):
    """Traffic direction scenario"""

    DOWNLOAD = 0
    """Server transmits payload to client."""
    UPLOAD = 1
    """Client transmits payload to server."""
    BOTH = 2
    """Payload is transmitted in both directions."""
    ECHO = 3
    """Client transmits payload to server, server echoes the payload back"""


class PayloadGenerationMethod(IntEnum):
    """Payload generation method."""

    FIXED = 0
    """Payload has a fixed value."""
    INCREMENT = 1
    """Payload consist of incrementing bytes."""
    RANDOM = 2
    """Payload consists of pseudo random bytes with a repeat cycle of 1 MB."""
    LONGRANDOM = 3
    """Payload consists of pseudo random bytes with a repeat cycle of 4 GB."""


class InfiniteOrFinite(IntEnum):
    INFINITE = 0
    FINITE = 1


class WhoClose(IntEnum):
    """How to close TCP connection when all payload has been transmitted."""

    NONE = 0
    """Keep the connection open after last byte is transmitted"""
    CLIENT = 1
    """Client closes the connection after last byte is receiver/transmitted"""
    SERVER = 2
    """Server closes the connection after last byte is transmitted"""


class LifecycleMode(IntEnum):
    """Connection lifecycle mode"""

    ONCE = 0
    """Connections are established during the ramp-up phase and not closed until the ramp-down phase of the load profile. That is, each configured connection only exists once."""
    IMMORTAL = 1
    """Connections are established during the ramp-up phase of the load profile, and are closed after the configured lifetime (configured by  P4G_RAW_CONN_LIFETIME). As connections close, new connections are established, attempting to keep the concurrent number of established connections constant. A new connection will have the same IP address as the connection it replaces, but will have a new TCP port number. This will simulate that the user (defined by the client IP address) is living on, and creates new connections as old connections close."""
    REINCARNATE = 2
    """Connections are established during the ramp-up phase of the load profile, and are closed after the configured lifetime (configured by  P4G_RAW_CONN_LIFETIME). As connections close, new connections are established, attempting to keep the concurrent number of established connections constant. A new connection will have the same TCP port number as the connection it replaces, but will have a new IP address. This will simulate that the user (defined by the client IP address) ceases to exist, and new users appear as old users die."""


class L47IPVersion(IntEnum):
    """IP version of the Connection Group"""

    IPV4 = 4
    """IPv4"""
    IPV6 = 6
    """IPv6"""


class L47ProtocolType(IntEnum):
    """L4 protocol of the Connection Group"""

    TCP = 0
    """TCP"""
    UDP = 1
    """UDP"""


class L47TrafficState(IntEnum):
    """L47 traffic state"""

    OFF = 0
    """Off"""
    ON = 1
    """On"""
    STOP = 2
    """Stop"""
    PREPARE = 3
    """Prepare"""
    PRERUN = 4
    """Prerun"""


class L47PortState(IntEnum):
    """L47 port state"""

    OFF = 0
    """Off"""
    PREPARE = 1
    """Prepare"""
    PREPARE_RDY = 2
    """Prepare Ready"""
    PREPARE_FAIL = 3
    """Prepare Failed"""
    PRERUN = 4
    """Prerun"""
    PRERUN_RDY = 5
    """Prepare Ready"""
    RUNNING = 6
    """Running"""
    STOPPING = 7
    """Stopping"""
    STOPPED = 8
    """Stopped"""


class L47PortSpeed(IntEnum):
    """L47 port speed mode"""

    AUTO = 0
    """Auto"""
    F100M = 1
    """100 Mbit/s"""
    F1G = 2
    """1 Gbit/s"""
    F2_5G = 3
    """2.5 Gbit/s"""
    F5G = 4
    """5 Gbit/s"""
    F10G = 5
    """10 Gbit/s"""
    F25G = 6
    """25 Gbit/s"""
    F40G = 7
    """40 Gbit/s"""
    F50G = 8
    """50 Gbit/s"""
    F100G = 9
    """100 Gbit/s"""


class CaptureSize(IntEnum):
    """Capture size"""

    FULL = 0
    """Capture whole packets"""
    SMALL = 1
    """Capture truncated packets"""


class ReplayParserState(IntEnum):
    OFF = 0
    PARSING = 1


class IsPresent(IntEnum):
    NOT_PRESENT = 0
    PRESENT = 1


class LicenseSpeed(IntEnum):
    UNDEFINED = 0
    F1G = 1
    F2_5G = 2
    F5G = 3
    F10G = 4
    F25G = 5
    F40G = 6
    F50G = 7
    F100G = 8


class TLSVersion(IntEnum):
    """TLS protocol version"""

    SSLV3 = 0
    """SSL v3"""
    TLS10 = 1
    """TLS v1.0"""
    TLS11 = 2
    """TLS v1.1"""
    TLS12 = 3
    """TLS v1.2"""


class ResourceAllocationMode(IntEnum):
    SIMPLE = 0
    ADVANCED = 1


class ReplaySchedulingMode(IntEnum):
    BANDWIDTH = 0
    TIME = 1


class ReplaySyncBasedOn(IntEnum):
    PER_CONN = 0
    PER_USER = 1


# endregion

# region Impairment enums
class CorruptionType(IntEnum):
    """Impairment corruption type"""

    OFF = 0
    """Off"""
    ETH = 1
    """Ethernet"""
    IP = 2
    """IP"""
    UDP = 3
    """UDP"""
    TCP = 4
    """TCP"""
    BER = 5
    """Bit Error Rate"""


class PolicerMode(IntEnum):
    """Policer mode"""

    L1 = 0
    """Policer performed at Layer 1 level. I.e. including the preamble and min inter-packet gap."""
    L2 = 1
    """Policer performed at Layer 2 level. I.e. excluding the preamble and min inter-packet gap"""


class FilterUse(IntEnum):
    """Use of filter."""

    OFF = 0
    """No filtering will be done"""
    AND = 1
    """Filtering will be done"""


class InfoAction(IntEnum):
    """Action of filter."""

    EXCLUDE = 0
    """Matching packets are excluded from the flow"""
    INCLUDE = 1
    """Matching packets are included from the flow"""


class L2PlusPresent(IntEnum):
    """Presence of Layer-2+ protocols"""

    NA = 0
    """No Layer 2+ protocols"""
    VLAN1 = 1
    """One VLAN Tag is present"""
    VLAN2 = 2
    """Two VLAN Tags are present"""
    MPLS = 3
    """MPLS label is present"""


class L3PlusPresent(IntEnum):
    """Presence of Layer-3 protocols"""

    NA = 0
    """No Layer 3 protocols"""
    IP4 = 1
    """IPv4 is present"""
    IP6 = 2
    """IPv6 is present"""


class FilterMode(IntEnum):
    """Impairment Filter Mode"""

    BASIC = 0
    """Basic Mode"""
    EXTENDED = 1
    """Extended Mode"""


class ImpairmentLatencyMode(IntEnum):
    """Impairment Latency Mode"""

    NORMAL = 0
    """Normal"""
    EXTENDED = 1
    """Extended"""


class ShadowWorkingSelection(IntEnum):
    """Shadow Working Selection"""

    SHADOW = 0
    """Shadow"""
    WORKING = 1
    """Working"""


class FilterType(IntEnum):
    """Filter Type for Impairment"""

    SHADOW = 0
    """Shadow Copy"""
    WORKING = 1
    """Working Copy"""


class FilterVlanType(IntEnum):
    """VLAN PCP Settings for VLAN Filter"""

    INNER = 0
    """VLAN1 (0) (INNER VLAN Tag is specified for the filter – used also when only 1 VLAN), indicates single/inner VLAN-TPID=0x8100"""
    OUTER = 1
    """VLAN2 (1) (OUTER VLAN Tag is specified for the filter), indicates outer VLAN-TPID=0x88A8"""


class LatencyTypeCustomDist(IntEnum):
    """Latency Type for Custom Distribution"""

    INTERPACKET_DISTRIBUTION = 0
    """Interpacket Distribution"""
    LATENCY_DISTRIBUTION = 1
    """Latency Distribution"""


class ImpairmentTypeIndex(IntEnum):
    """Impairment Type Index"""

    DROP = 0
    """Drop"""
    MISORDER = 1
    """Misorder"""
    LATENCYJITTER = 2
    """Delay/Jitter"""
    DUPLICATION = 3
    """Duplication"""
    CORRUPTION = 4
    """Corruption"""
    POLICER = 5
    """Policer"""
    SHAPER = 6
    """Shaper"""


# endregion

# region TSN enums
class TSNConfigProfile(IntEnum):
    """TSN PTP Configuration profile"""

    AUTOMOTIVE = 0
    """Defaults suitable for automotive testing"""
    IEEE1588V2 = 1
    """Defaults suitable for PTP  testing"""


class TSNPortRole(IntEnum):
    """TSN port role"""

    GRANDMASTER = 0
    """Grandmaster role"""
    SLAVE = 1
    """Slave role"""


class TSNDeviationMode(IntEnum):
    FIXED = 0


class TSNTimeSource(IntEnum):
    """TSN time source"""

    ATOMIC = 0x10
    """Atomic"""
    GPS = 0x20
    """GPS"""
    TERRESTRIAL = 0x30
    """Terrestrial"""
    PTP = 0x40
    """PTP"""
    NTP = 0x50
    """NTP"""
    HAND_SET = 0x60
    """Handset"""
    OTHER = 0x90
    """Other"""
    INTERNAL_OSC = 0xA0
    """Internal oscillator"""


class TSNHistogramSource(IntEnum):
    """Data source for TSN histogram"""

    DRIFT = 0
    """Post-servo offset to Grandmaster"""
    DRIFTPRE = 1
    """Pre-servo offset to Grandmaster"""
    PDELAY = 2
    """PDelay data."""
    NRR = 3
    """Neighbor Rate Ratio data."""


class TSNStatisticsTypes(IntEnum):
    """TSN Statistics Types"""

    ALL = 0
    """All."""
    PACKETCOUNT = 1
    """Packet count."""
    OFFSET = 2
    """Offset."""
    PDELAY = 3
    """PDelay."""
    SYNCRATE = 4
    """Sync rate."""


# endregion

# region ANLT enums


class AutoNegMode(IntEnum):
    """Auto Neg Mode"""

    ANEG_OFF = 0
    """Auto Neg Off"""
    ANEG_ON = 1
    """Auto Neg On"""


class AutoNegTecAbility(IntEnum):
    """Auto Neg Technical Abilities"""

    DEFAULT_TECH_MODE = 0
    """Default Tech Mode"""
    IEEE_10G_KR = 4
    """IEEE 10G KR"""
    IEEE_40G_CR4 = 16
    """IEEE 40G CR4"""
    IEEE_100G_KR4 = 128
    """IEEE 100G KR4"""
    IEEE_100G_CR4 = 256
    """IEEE 100G CR4"""
    IEEE_25GBASE_CRS_KRS = 512
    """IEEE 25GBASE CRS KRS"""
    IEEE_25GBASE_CR_KR = 1024
    """IEEE 25GBASE CR KR"""
    IEEE_50GBASE_CR_KR = 8192
    """IEEE 50GBASE CR KR"""
    IEEE_100GBASE_CR2_KR2 = 16384
    """IEEE 100GBASE CR2 KR2"""
    IEEE_200GBASE_CR4_KR4 = 32768
    """IEEE 200GBASE CR4 KR4"""
    IEEE_100GBASE_KR1 = 65536
    """IEEE 100GBASE KR1"""
    IEEE_200GBASE_KR2 = 131072
    """IEEE 200GBASE KR2"""
    IEEE_400GBASE_KR4 = 262144
    """IEEE 400GBASE KR4"""
    EC_25GBASE_KR1 = 16777216
    """EC 25GBASE KR1"""
    EC_25GBASE_CR1 = 33554432
    """EC 25GBASE CR1"""
    EC_50GBASE_KR2 = 67108864
    """EC 50GBASE KR2"""
    EC_50GBASE_CR2 = 134217728
    """EC 50GBASE CR2"""
    EC_400GBASE_KR8 = 268435456
    """EC 400GBASE KR8"""
    EC_800GBASE_KR8 = 536870912
    """EC 800GBASE KR8"""
    EC_50G_CR1_KR1 = 503
    """EC 50G CR1 KR1"""
    BAM_50G_CR1_KR1 = 504
    """BAM 50G CR1 KR1"""
    BAM_50G_CR2_KR2 = 505
    """BAM 50G CR2 KR2"""
    BAM_100G_CR2_KR2 = 1002
    """BAM 100G CR2 KR2"""
    BAM_100G_CR4_KR4 = 1003
    """BAM 100G CR4 KR4"""
    BAM_200G_CR2_KR2 = 2002
    """BAM 200G CR2 KR2"""
    BAM_400G_CR8_KR8 = 4001
    """BAM 400G CR8 KR8"""


class AutoNegFECOption(IntEnum):
    """Auto Neg FEC Options"""

    DEFAULT_FEC = 0
    """Default FEC"""
    NO_FEC = 1
    """No FEC"""
    FCFEC = 2
    """Firecode FEC"""
    RSFEC_CL91 = 4
    """RS FEC Cl91"""
    RS528 = 256
    """RS 528"""
    RS544 = 512
    """RS 544"""
    RS272 = 1024
    """RS 272"""


class AutoNegFECType(IntEnum):
    """Auto Neg FEC Type"""

    PENDING = 0
    """Pending"""
    NO_FEC = 1
    """No FEC"""
    RS_FEC = 513
    """RS FEC"""
    FC_FEC = 257
    """Firecode FEC"""


class AutoNegStatus(IntEnum):
    """Auto Neg Status"""

    UNKNOWN = 0
    """Unknown"""
    ENABLE = 1
    """Enabled"""
    TRANSMIT_DISABLE = 2
    """Transmit Disabled"""
    ABILITY_DETECT = 3
    """Ability Detected"""
    ACKNOWLEDGE_DETECT = 4
    """Acknowledge Detected"""
    COMPLETE_ACKNOWLEDGE = 5
    """Complete Acknowledge"""
    NEXT_PAGE_WAIT = 6
    """Next Page Wait"""
    AN_GOOD_CHECK = 7
    """AN Good Check"""
    AN_GOOD = 8
    """AN Good"""


class AutoNegFECStatus(IntEnum):
    """Auto Neg FEC Status"""

    DEFAULT_FEC = 0
    """Default FEC"""
    NO_FEC = 1
    """No FEC"""
    FC_FEC = 2
    """Firecode FEC"""
    RSFEC_CL91 = 4
    """RS FEC Cl91"""
    RS528 = 256
    """RS 528"""
    RS544 = 512
    """RS 544"""
    RS272 = 1024
    """RS 272"""


class LinkTrainingMode(IntEnum):
    """Link Training Mode"""

    START_AFTER_AUTONEG = 0
    """Link training starts automatically after autoneg is completed"""

    STANDALONE = 1
    """Link training procedure is done automatically by the port"""

    DISABLED = 2
    """Link training is disabled"""

    INTERACTIVE = 3
    """Link training in interactive mode, requiring manual operation."""

    UNKNOWN = 255
    """Unknown link training mode"""


class PAM4FrameSize(IntEnum):
    """PAM4 Frame Size"""

    P16K_FRAME = 0
    """16K Frame Size"""
    P4K_FRAME = 1
    """4K Frame Size"""


class LinkTrainingInitCondition(IntEnum):
    """Link Training Initialization Condition"""

    NO_INIT = 0
    """No Initialization"""
    INIT_ENABLED = 1
    """Initialization Enabled"""


class NRZPreset(IntEnum):
    """Link Training NRZ Preset"""

    NRZ_NO_PRESET = 0
    """NRZ without Preset"""
    NRZ_WITH_PRESET = 1
    """NRZ with Preset"""


class TimeoutMode(IntEnum):
    """Link Training Timeout Mode"""

    DEFAULT = 0
    """Default Timeout"""
    DISABLED = 255
    """Timeout Disabled"""


class LinkTrainingStatusMode(IntEnum):
    """Link Training Status Mode"""

    DISABLED = 0
    """Disabled"""
    ENABLED = 1
    """Enabled"""


class LinkTrainingStatus(IntEnum):
    """Link Training Status"""

    NOT_TRAINED = 0
    """Not Trained"""
    TRAINED = 1
    """Trained"""


class LinkTrainingFailureType(IntEnum):
    """Link Training Failure Type"""

    NO_FAILURE = 0
    """No Failure"""
    FRAME_LOCK_ERROR = 1
    """Frame Lock Error"""
    SNR_BELOW_THRESHOLD = 2
    """SNR Below Threshold"""
    TIME_OUT_FAILURE = 3
    """Timeout Failure"""


class Layer1ConfigType(IntEnum):
    """
    .. versionadded:: 1.1

    Enums for PL1_CFG_TMP's type.
    """

    ANLT_INTERACTIVE = 0
    """ANLT Interactive mode config"""

    AN_LOOPBACK = 1
    """Auto-negotiation loopback config"""

    LT_INITIAL_MODULATION = 2
    """The initial modulation (0=NRZ, 1=PAM4, 2=PAM4_WITH_PRECODING) """

    LL_DEBUG_INFO = 3
    """Return the an/lt module base and RX and TX (serdes index, base address)"""

    LT_TRAINING_ALGORITHM = 4
    """The link training algorithm to use"""

    ANLT_LOG_CONTROL = 5
    """Control what should be logged by anlt"""

    ANLT_STRICT_MODE = 6
    """Set AN/LT strict mode. In strict mode errored framed will be ignored"""

    AN_LT_XLA_MODE = 7
    """Set XLA mode. If enabled XLA dumps will, if triggered, be logged automatically"""



class Layer1LogType(IntEnum):
    """

    .. versionadded:: 1.1

    .. warning::

        Still in beta mode. Subjected to changes

    Enums for PL1_LOG's type.
    """

    AN = 0
    """Log for auto-neg"""

    LT = 1
    """Log for link training"""


class LinkTrainAlgorithm(IntEnum):
    """
    .. versionadded:: 1.2

    Link Training Algorithm

    """

    INTERACTIVE = 0
    """INTERACTIVE"""

    ALG0 = 1
    """ALGORITHM 0"""

    ALGN1 = 2
    """ALGORITHM -1"""


class LinkTrainCmd(IntEnum):
    """
    .. versionadded:: 1.1

    Link Training commands

    """

    CMD_NOP = 0
    """No operation. Used for 'ping' testing"""

    CMD_INC = 1
    """Increment the coeff provided in ARG"""

    CMD_DEC = 2
    """Decrement the coeff provided in ARG"""

    CMD_PRESET = 3
    """Set the preset provided in ARG"""

    CMD_ENCODING = 4
    """Set encoding provided in ARG"""

    CMD_LOCAL_TRAINED = 255
    """Signal training completed"""


class LinkTrainPresets(IntEnum):
    """
    .. versionadded:: 1.1

    Link Training presets

    """

    PRESET_1 = 0
    """Preset 1"""

    PRESET_2 = 1
    """Preset 2"""

    PRESET_3 = 2
    """Preset 3"""

    PRESET_4 = 3
    """Preset 4"""

    PRESET_5 = 4
    """Preset 5"""

    UNKNOWN = 255
    """Unknown coeff"""


class LinkTrainCoeffs(IntEnum):
    """
    .. versionadded:: 1.1


    Link Training coefficients

    """

    PRE = 0
    """Pre coeff c(-1)"""

    MAIN = 1
    """Main coeff c(0)"""

    POST = 2
    """Post coeff c(1)"""

    PRE2 = 3
    """Pre2 coeff c(-2)"""

    PRE3 = 4
    """Pre3 coeff c(-3)"""

    UNKNOWN = 255
    """Unknown coeff"""


class LinkTrainEncoding(IntEnum):
    """
    .. versionadded:: 1.1

    Link Training Encoding

    """

    NRZ = 0
    """NRZ (PAM2)"""

    PAM4 = 1
    """PAM4"""

    PAM4_WITH_PRECODING = 2
    """PAM4 with precoding"""

    UNKNOWN = 255
    """PAM4 with precoding"""


class LinkTrainCmdResults(IntEnum):
    """
    .. versionadded:: 1.1

    Link Training Command Results

    """

    UNKNOWN = 0x00 | 0
    """Unknown result"""

    SUCCESS = 0x00 | 1
    """Command successfully completed"""

    TIMEOUT = 0x00 | 2
    """Command timeout"""

    FAILED = 0x00 | 3
    """Command failed"""

    COEFF_STS_NOT_UPDATED = 0x80 | 0
    """Coeff did not update"""

    COEFF_STS_UPDATED = 0x80 | 1
    """Coeff updated"""

    COEFF_STS_AT_LIMIT = 0x80 | 2
    """Coeff at limit"""

    COEFF_STS_NOT_SUPPORTED = 0x80 | 3
    """Coeff not supported"""

    COEFF_STS_EQ_LIMIT = 0x80 | 4
    """EQ limit reached"""

    COEFF_STS_C_AND_EQ_LIMIT = 0x80 | 6
    """Coeff and EQ limit reached"""


class LinkTrainCmdFlags(IntEnum):
    """
    .. versionadded:: 1.1

    Link Training Command Flags

    """

    NEW = 1
    """New command"""

    IN_PROGRESS = 2
    """Command in progress"""

    DONE = 4
    """Command done"""

    LOCK = 8
    """Link locked"""

    LOCK_LOST = 16
    """Link lock lost"""

    OVERRUN = 32
    """Overrun detected"""

class LinkTrainAnnounce(IntEnum):
    """
    .. versionadded:: 1.1

    Link Training Announce

    """

    TRAINED = 0
    """The lane is trained"""


class AnLtLogControl(IntEnum):
    """
    .. versionadded:: 1.3

    ANLT log control bits

    """
    # 1st nibble
    LOG_TYPE_DEBUG = 0x2
    """debug log output"""

    LOG_TYPE_AN_TRACE = 0x4
    """autonegotiation trace output"""

    LOG_TYPE_LT_TRACE = 0x8
    """link training trace output"""

    # 2nd nibble
    LOG_TYPE_ALG_TRACE = 0x10
    """link training algorithm trace"""

    # 5th nibble
    LOG_TYPE_FSM_PORT = 0x10000
    """port state machine transitions"""

    LOG_TYPE_FSM_ANEG = 0x20000
    """autonegotiation state machine transitions. What we act on"""

    LOG_TYPE_FSM_ANEG_STIMULI = 0x40000
    """autonegotiation stimuli state machine transitions. What we ask"""

    LOG_TYPE_FSM_LT = 0x80000
    """link training state machine transitions"""

    # 6th nibble
    LOG_TYPE_FSM_LT_COEFF = 0x100000
    """link training coefficient state machine transitions. What we act on"""

    LOG_TYPE_FSM_LT_STIMULI = 0x200000
    """link training stimuli state machine transitions. What we ask"""

    LOG_TYPE_FSM_LT_ALG0 = 0x400000
    """link training algorithm 0 state machine transitions"""

    LOG_TYPE_FSM_LT_ALG1 = 0x800000
    """link training algorithm -1 state machine transitions"""

# endregion



import typing
import asyncio
import functools
from typing import TYPE_CHECKING
from typing_extensions import Self
from xoa_driver.internals.core.commands import (
    M_STATUS,
    M_UPGRADE,
    M_UPGRADEPROGRESS,
    M_CFPTYPE,
    M_COMMENT,
    M_CAPABILITIES,
    M_CLOCKPPB,
    M_TXCLOCKSOURCE_NEW,
    M_TXCLOCKSTATUS_NEW,
    M_EMULBYPASS,
    M_LATENCYMODE,
    M_REVISION,
    M_MEDIA,
    M_MEDIASUPPORT,
    M_TIMESYNC,
    M_CLOCKSYNCSTATUS,
    M_NAME,
    M_CFPCONFIGEXT,
    M_CFPCONFIG,
)

from xoa_driver.internals.hli_v1 import revisions
from xoa_driver.internals.utils import ports_manager as pm
from xoa_driver.internals.utils import attributes as utils
from xoa_driver.internals.state_storage import modules_state
from xoa_driver import ports
from . import base_module as bm
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
    from . import __interfaces as m_itf


class ChTXClock:
    """
    Advanced timing feature (Chimera).
    """
    def __init__(self, conn: "itf.IConnection", module_id: int) -> None:
        self.source = M_TXCLOCKSOURCE_NEW(conn, module_id)
        """
        The source that drives the TX clock rate of the ports on the test module.
        
        :type: M_TXCLOCKSOURCE_NEW
        """
        self.status = M_TXCLOCKSTATUS_NEW(conn, module_id)
        """
        TX clock status of the test module.
        
        :type: M_TXCLOCKSTATUS_NEW
        """


class ChCFP:
    """
    CFP test module (Chimera).
    """
    def __init__(self, conn: "itf.IConnection", module_id: int) -> None:
        self.type = M_CFPTYPE(conn, module_id)
        """
        The transceiver's CFP type currently inserted.
        
        :type: M_CFPTYPE
        """
        self.config = M_CFPCONFIG(conn, module_id)
        """
        The CFP configuration of the test module.
        
        :type: M_CFPCONFIG
        """
        self.config_extended = M_CFPCONFIGEXT(conn, module_id)
        """
        The CFP configuration of the test module.
        
        :type: M_CFPCONFIGEXT
        """


class ChUpgrade:
    """
    Upgrade test module (Chimera).
    """
    def __init__(self, conn: "itf.IConnection", module_id: int) -> None:
        self.start = M_UPGRADE(conn, module_id)
        """
        Start the upgrade progress of the test module.
        
        :type: M_UPGRADE
        """
        self.progress = M_UPGRADEPROGRESS(conn, module_id)
        """
        Upgrade progress status of the test module.
        
        :type: M_UPGRADEPROGRESS
        """

class ChTiming:
    """Test module timing and clock configuration"""

    def __init__(self, conn: "itf.IConnection", module_id: int) -> None:
        self.source = M_TIMESYNC(conn, module_id)
        """Timing source of the test module.

        :type: M_TIMESYNC
        """

        self.clock_local_adjust = M_CLOCKPPB(conn, module_id)
        """Time adjustment controlling of the local clock of the test module, which drives the TX rate of the test ports.
        
        :type: M_CLOCKPPB
        """

        self.clock_sync_status = M_CLOCKSYNCSTATUS(conn, module_id)
        """Test module's clock sync status.

        :type: M_CLOCKSYNCSTATUS
        """


class ModuleChimera(bm.BaseModule["modules_state.ModuleLocalState"]):
    """
    This is a conceptual class of Chimera module.
    """
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)

        self._local_states = modules_state.ModuleLocalState()

        self.tx_clock = ChTXClock(conn, self.module_id)
        """
        TX clock config (Chimera).

        :type: ChTXClock
        """

        self.timing = ChTiming(conn, self.module_id)
        """
        Timing config (Chimera).

        :type: ChTiming
        """

        self.cfp = ChCFP(conn, self.module_id)
        """
        CFP test module (Chimera).

        :type:  ChCFP
        """

        self.upgrade = ChUpgrade(conn, self.module_id)
        """
        Upgrade test module (Chimera).

        :type: ChUpgrade
        """

        self.capabilities = M_CAPABILITIES(conn, self.module_id)
        """
        Test module's capabilities.
        
        :type: M_CAPABILITIES
        """

        self.comment = M_COMMENT(conn, self.module_id)
        """
        Test module's description.
        
        :type: M_COMMENT
        """

        self.status = M_STATUS(conn, self.module_id)
        """
        Test module's status.
        
        :type: M_STATUS
        """

        self.clock_ppb = M_CLOCKPPB(conn, self.module_id)
        """
        Test module's local clock adjustment.
        
        :type: M_CLOCKPPB
        """

        self.emulator_bypass_mode = M_EMULBYPASS(conn, self.module_id)
        """
        Bypass mode of the Chimera module.
        
        :type: M_EMULBYPASS
        """

        self.latency_mode = M_LATENCYMODE(conn, self.module_id)
        """
        Latency mode of the Chimera module.
        
        :type: M_LATENCYMODE
        """

        self.revision = M_REVISION(conn, self.module_id)
        """Test module's model P/N name.

        :type: M_REVISION
        """

        self.media = M_MEDIA(conn, self.module_id)
        """Test module's media type.

        :type: M_MEDIA
        """

        self.available_speeds = M_MEDIASUPPORT(conn, self.module_id)
        """Test module's available speeds.

        :type: M_MEDIASUPPORT
        """

        self.name = M_NAME(conn, self.module_id)
        """Test module's name.

        :type: M_NAME
        """

        self.ports: pm.PortsManager["ports.PortChimera"] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PortChimera,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """
        Port index manager of the Chimera module.

        :type: PortsManager
        """

    @property
    def info(self) -> modules_state.ModuleLocalState:
        return self._local_states

    async def _setup(self) -> Self:
        await asyncio.gather(
            self._local_states.initiate(self),
            self.ports.fill()
        )
        self._local_states.register_subscriptions(self)
        return self

    on_cfp_type_change = functools.partialmethod(utils.on_event, M_CFPTYPE)
    """
    Register a callback function to the event that the module's CFP type changes.
    """

    on_cfp_config_change = functools.partialmethod(utils.on_event, M_CFPCONFIG)
    """
    Register a callback function to the event that the module's CFP configuration changes.
    """

    on_cfp_config_extended_change = functools.partialmethod(utils.on_event, M_CFPCONFIGEXT)
    """
    Register a callback function to the event that the module's CFP configuration changes.
    """

    on_status_change = functools.partialmethod(utils.on_event, M_STATUS)
    """
    Register a callback function to the event that the module's model changes.
    """

    on_latency_mode_change = functools.partialmethod(utils.on_event, M_LATENCYMODE)
    """
    Register a callback function to the event that the module's latency mode changes.
    """

    on_media_change = functools.partialmethod(utils.on_event, M_MEDIA)
    """
    Register a callback to the event that the module's media and available speeds change.
    """

    on_media_support_change = functools.partialmethod(utils.on_event, M_MEDIASUPPORT)
    """
    Register a callback to the event that the module's supported media changes.
    """

    on_timing_source_change = functools.partialmethod(utils.on_event, M_TIMESYNC)
    """
    Register a callback to the event that the module's timesync mode changes.
    """

    on_timing_clock_local_adjust_change = functools.partialmethod(utils.on_event, M_CLOCKPPB)
    """
    Register a callback to the event that the module's clock adjustment ppb changes.
    """

    on_adv_timing_clock_tx_status_change = functools.partialmethod(utils.on_event, M_TXCLOCKSTATUS_NEW)
    """Register a callback to the event that the module's TX clock status changes."""

    on_adv_timing_clock_tx_source_change = functools.partialmethod(utils.on_event, M_TXCLOCKSOURCE_NEW)
    """Register a callback to the event that the module's clock that drives the port TX rates changes."""


@typing.final
@revisions.register_chimera_module(rev="Chimera-100G-5S-2P")
class MChi100G5S2P(ModuleChimera):
    """Chimera module Chi-100G-5S-2P"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PChi100G5S2P] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PChi100G5S2P,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port index manager of Chi-100G-5S-2P
        
        :type: PortsManager
        """


@typing.final
@revisions.register_chimera_module(rev="Chimera-100G-5S-2P[b]")
class MChi100G5S2P_b(ModuleChimera):
    """Chimera module Chi-100G-5S-2P[b]"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PChi100G5S2P_b] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PChi100G5S2P_b,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port index manager of Chi-100G-5S-2P[b]
        
        :type: PortsManager
        """


@typing.final
@revisions.register_chimera_module(rev="Chimera-40G-2S-2P")
class MChi40G2S2P(ModuleChimera):
    """Chimera module Chi-40G-2S-2P"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PChi40G2S2P] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PChi40G2S2P,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port index manager of Chi-40G-2S-2P

        :type: PortsManager
        """



from functools import partial
from xoa_driver.internals.utils.rev_tool import RegisterModule

VALKYRIE_MODULES = {}
CHIMERA_MODULES = {}
VULCAN_MODULES = {}


register_valkyrie_module = partial(RegisterModule, modules_store=VALKYRIE_MODULES)
register_chimera_module = partial(RegisterModule, modules_store=CHIMERA_MODULES)
register_vulcan_module = partial(RegisterModule, modules_store=VULCAN_MODULES)



from enum import IntEnum
from typing import (
    Any,
    Generic,
    List,
    Tuple,
    Optional,
    Type,
    Union,
    NamedTuple,
)
from collections import UserList
from . import exceptions as excp
from . import interfaces as itf


def patch_type(base_type: Type, choices: Type["IntEnum"], bit_map: List[str]) -> Type:
    """Function dynamically patch XTypes for show meaningful choices value"""
    def _choice_str(self) -> str:

        def c_format(cho, itm) -> str:
            return f"{cho.__name__}.{cho(int(itm)).name}"

        if issubclass(base_type, UserList):
            return str([c_format(choices, i) for i in self])
        return c_format(choices, self)

    # def _bit_map_str(self) -> str:
    #     results = []
    #     for e in self:
    #         one_result = [b for i, b in enumerate(bit_map) if e & (1 << i)]
    #         results.append(one_result)
    #     return str(results)

    dic = {}
    if choices:
        dic["__str__"] = _choice_str
        dic["__repr__"] = _choice_str
    # elif bit_map: # need to improve logic this part will not work coz function will be colled only if <choices> are defined
    #     dic["__str__"] = _bit_map_str
    #     dic["__repr__"] = _bit_map_str
    return type(base_type.__name__, (base_type,), dic)


class ClimbRange(NamedTuple):
    min: Union[int, float]
    max: Union[int, float]


class XmpField(Generic[itf.XmpGenericField]):
    """
    Xmp Field descriptor for command structures.
    Taking care of all passed values to the Packet structure will be
    in correct convertable type for Xena Management Protocol.
    """

    def __init__(
        self,
        xmp_type: Type[itf.XmpGenericField],
        *,
        choices: Optional[Type["IntEnum"]] = None,
        climb: Optional[Tuple[Union[int, float], Union[int, float]]] = None,
    ) -> None:
        self.xmp_type = xmp_type

        self.choices: Optional[Type["IntEnum"]] = None
        self.climb: Optional[ClimbRange] = None
        self.bit_map: List[str] = []

        is_numerical = issubclass(self.xmp_type, (int, float))

        if is_numerical or issubclass(self.xmp_type, UserList):
            if choices:
                self.choices = choices
                self.xmp_type = patch_type(
                    self.xmp_type,
                    self.choices,
                    self.bit_map
                )
        if is_numerical:
            self.climb = ClimbRange(*climb) if climb else None

    def __validate_choices(self, val: "itf.XmpGenericField") -> None:
        if not self.choices:
            return None
        elif isinstance(val, UserList):
            [self.choices(v) for v in val]
        else:
            self.choices(val)  # raise an exception if value is not in provided enum

    def __validate_limits(self, val: Union[int, float]) -> None:
        if not self.climb:
            return None
        elif not (self.climb.min <= val <= self.climb.max):
            raise excp.ValueOutOfRangeError(val, self.climb)

    def __set_name__(self, owner: object, name: str) -> None:
        self._private_name = name

    def __set__(self, obj: object, value: Any) -> None:
        val = (
            self.xmp_type.from_bytes(value)
            if isinstance(value, bytes)
            else self.xmp_type(value)
        )

        self.__validate_choices(val)  # type: ignore
        self.__validate_limits(val)  # type: ignore
        obj.__dict__[self._private_name] = val

    def __get__(self, obj: object, objtype: Any = None) -> "itf.XmpGenericField":
        return obj.__dict__[self._private_name]


# class StateStringType:
#     @staticmethod
#     def validate(inst, val) -> None:
#         return None

# class StateNumericalType:
#     @staticmethod
#     def _validate_choices(inst, val) -> None:
#         if not inst.choices:
#             return None
#         inst.choices(val)  # raise an exception if value is not in provided enum

#     @staticmethod
#     def _validate_limits(inst, val: Union[int, float]) -> None:
#         if not inst.climb:
#             return None
#         elif not (inst.climb.min <= val <= inst.climb.max):
#             raise excp.ValueOutOfRangeError(val, inst.climb)

#     @staticmethod
#     def validate(inst, val) -> None:
#         StateNumericalType._validate_choices(inst, val)
#         StateNumericalType._validate_limits(inst, val)


# class StateIterableNumericalType:
#     @staticmethod
#     def _validate_choices(inst, val) -> None:
#         if not inst.choices:
#             return None
#         [ inst.choices(v) for v in val ]

#     @staticmethod
#     def _validate_limits(inst, val: Union[List[int], List[float]]) -> None:
#         if not inst.climb:
#             return None
#         elif any((inst.climb.min > v > inst.climb.max) for v in val):
#             raise excp.ValueOutOfRangeError(val, inst.climb)

#     @staticmethod
#     def validate(inst, val) -> None:
#         StateIterableNumericalType._validate_choices(inst, val)
#         StateIterableNumericalType._validate_limits(inst, val)



import asyncio
from concurrent.futures import ThreadPoolExecutor
from asyncio.futures import Future
from typing import (
    Callable,
    TYPE_CHECKING,
    Optional,
    Tuple,
    Type,
    Union,
)

from uuid import uuid4

if TYPE_CHECKING:
    from asyncio.transports import Transport

from ..protocol import command_builders
from ..protocol import constants as const
from ..registry import COMMANDS_REGISTRY
from .. import interfaces as x_types
from .. import protocol

from . import events_observer as eveo
from . import logging as log
from . import exceptions as t_ex
from . import commands_manager as cm
from .request_id import RequestId


class TransportationHandler(asyncio.Protocol):
    """Handling sending and receiving of the XMP commands."""

    __slots__ = (
        "identity",
        "is_connected",
        "__log",
        "__transport",
        "events_observer",
        "cmd_mngr",
        "__req_id",
        "commands_registry",
        "magic_word",
        "aggregate",
        "_worker",
    )

    def __init__(self, debug: bool = False) -> None:
        self.identity = uuid4().hex
        self.is_connected: bool = False
        self._worker = ThreadPoolExecutor()

        self.__log = log.TransportationLogger(self.identity[:6], debug)

        self.__transport: Optional["Transport"] = None
        self.events_observer = eveo.EventsObserver()
        self.cmd_mngr = cm.CommandsManager()
        self.__req_id: RequestId = RequestId()

        self.commands_registry = COMMANDS_REGISTRY
        self.magic_word = const.MAGIC_WORD
        self.aggregate: bytearray = bytearray()

    def connection_made(self, transport: "Transport") -> None:
        self.__transport = transport
        self.is_connected = True
        peername = transport.get_extra_info("peername")
        self.__log.info(f"Connection to {peername}")

    def __handle_push_response(self, response: protocol.Response) -> None:
        self.__log.push_obj(response)
        self.events_observer.dispatch(
            response.header.cmd_code,
            response
        )

    def __handle_param_response(self, response: protocol.Response) -> None:
        self.__log.response_obj(response)
        future = self.cmd_mngr.get_result_future(response)
        if not future:
            raise t_ex.LostFuture(response)
        if not response.is_ok:
            future.set_exception(t_ex.BadStatus(response))
        else:
            future.set_result(response.values)

    def data_received(self, data: bytes) -> None:
        """Process received data from xenaserver."""

        self.aggregate.extend(data)
        while len(self.aggregate) > 0:
            packet_position = self.read_data_with_magic_word(self.aggregate)
            if not packet_position:
                return None
            del self.aggregate[packet_position]
        # self.__log.draw_separator()

    def read_data_with_magic_word(self, extractor: bytearray) -> Optional[slice]:
        # read data which starts from magic word.
        HEADER_POS = slice(16)
        header_bytes = extractor[HEADER_POS]
        if (len(header_bytes) < HEADER_POS.stop) or (not header_bytes.startswith(self.magic_word)):
            return None
        header = protocol.ResponseHeader.from_buffer_copy(header_bytes)
        packet_body_length = header.body_size
        BODY_POS = slice(HEADER_POS.stop, HEADER_POS.stop + header.body_size)
        body_bytes = extractor[BODY_POS]
        if len(body_bytes) < packet_body_length:
            return None
        asyncio.create_task(
            self.process_data(header, body_bytes)
        )
        return slice(BODY_POS.stop)

    async def process_data(self, header, data: bytearray) -> None:
        try:
            response = await self.__serialize_to_response(header, data)
        except (t_ex.RepeatedRequestID, RuntimeError) as e:
            self.__log.error(f"{e} Original Data: {data}")
        else:
            if response.header.is_pushed:
                self.__handle_push_response(response)
            else:
                self.__handle_param_response(response)

    async def __serialize_to_response(self, header, data: bytearray) -> protocol.Response:
        """Applying received bytes to structured representation."""

        command_idx: Optional[int] = (
            header.cmd_code
            if header.is_pushed
            else self.cmd_mngr.get_command_handler_id(
                header.request_identifier
            )
        )
        if not command_idx:
            raise t_ex.RepeatedRequestID(header)
        xmc_type: Optional[Type[x_types.CMD_TYPE]] = self.commands_registry.get(command_idx, None)
        if not xmc_type:
            raise t_ex.NotImplementedCommand(header)
        r = self._worker.submit(
            command_builders.build_from_bytes,
            xmc_type,
            header,
            bytes(data)
        )
        return await asyncio.wrap_future(r)

    def eof_received(self) -> None:
        self.__log.info("EOF received")

    def connection_lost(self, exc) -> None:
        self.is_connected = False
        self._worker.shutdown()
        self.events_observer.dispatch(
            eveo.ON_EVT_DISCONNECTED,
            self.__transport.get_extra_info("peername") if self.__transport else None
        )
        self.__transport = None
        self.__log.info(f"The server closed the connection {exc}")

    def send(self, requests: Union[bytes, bytearray]) -> None:
        """
        Send applied commands from sending queue to
        xenaserver and liberate the sending queue.
        """

        if self.__transport is None:
            raise BrokenPipeError("No socket!")
        self.__transport.write(requests)
        # self.__log.draw_separator()

    def close(self) -> None:
        """Close connection with xenaserver."""
        if self.__transport is not None and not self.__transport.is_closing():
            self.__transport.close()

    async def prepare_data(self, request: protocol.Request) -> Tuple[bytes, Future]:
        assert self.is_connected, "Cannot add command because Socket is disconnected"
        request.header.request_identifier = await self.__req_id.get_avaliable()
        self.__log.request_obj(request)
        return bytes(request), self.cmd_mngr.append(request)

    def subscribe(self, xmc_cls: x_types.CMD_TYPE, callback: "Callable") -> None:
        assert xmc_cls.pushed, "Command is not subscribable."
        assert callback, "Callback function is required."
        self.events_observer.subscribe(
            xmc_cls.code,
            callback
        )

    def on_disconnected(self, callback: "Callable") -> None:
        self.events_observer.subscribe(
            eveo.ON_EVT_DISCONNECTED,
            callback
        )



from typing import (
    final,
    List,
    Type,
    TypeVar,
    TYPE_CHECKING,
)

from xoa_driver.internals.core.commands import (
    P4G_INDICES,
    P4G_CREATE,
    P4G_DELETE,
    P4G_ENABLE,
    P4G_COMMENT,
    P4G_CLEAR_COUNTERS,
    P4G_ROLE,
    P4G_LP_TIME_SCALE,
    P4G_LP_SHAPE,
    P4G_TEST_APPLICATION,
    P4G_L4_PROTOCOL,
)
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
    from xoa_driver.internals.utils import kind
from xoa_driver.internals.utils.indices import observer as idx_obs
from .tls import GTls
from .l2 import GL2
from .raw import GRaw
from .tcp import GTcp
from .udp import GUdp
from .replay import GReplay
from .l3 import GL3
from .user_state import GUserState
from .histogram import GHistogram

from ..base_index import BaseIndex


class GCounters:
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.clear = P4G_CLEAR_COUNTERS(conn, module_id, port_id, group_idx)


class GLoadProfile:
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.time_scale = P4G_LP_TIME_SCALE(conn, module_id, port_id, group_idx)
        self.shape = P4G_LP_SHAPE(conn, module_id, port_id, group_idx)


CG = TypeVar("CG")


@final
class ConnectionGroupIdx(BaseIndex):
    """L47 Connection Group Index Manager"""
    def __init__(self, conn: "itf.IConnection", kind: "kind.IndicesKind", observer: "idx_obs.IndicesObserver") -> None:
        super().__init__(conn, kind, observer)

        self.comment = P4G_COMMENT(self._conn, *kind)
        """
        Representation of p4g_commands.P4G_COMMENT
        """
        self.status = P4G_ENABLE(self._conn, *kind)
        """
        Representation of p4g_commands.P4G_ENABLE
        """
        self.role = P4G_ROLE(self._conn, *kind)
        """
        Representation of p4g_commands.P4G_ROLE
        """
        self.layer4_protocol = P4G_L4_PROTOCOL(self._conn, *kind)
        """
        Representation of p4g_commands.P4G_L4_PROTOCOL
        """
        self.test_application = P4G_TEST_APPLICATION(self._conn, *kind)
        """
        Representation of p4g_commands.P4G_TEST_APPLICATION
        """

        self.tls = GTls(self._conn, *kind)
        """TLS configurations."""
        self.l2 = GL2(self._conn, *kind)
        """L2 configurations."""
        self.raw = GRaw(self._conn, *kind)
        """Raw configurations."""
        self.tcp = GTcp(self._conn, *kind)
        """TCP configurations."""
        self.udp = GUdp(self._conn, *kind)
        """UDP configurations."""
        self.replay = GReplay(self._conn, *kind)
        """Replay configurations."""
        self.l3 = GL3(self._conn, *kind)
        """L3 configurations."""
        self.user_state = GUserState(self._conn, *kind)
        """User state configurations."""
        self.histogram = GHistogram(self._conn, *kind)
        """Histogram configurations."""
        self.counters = GCounters(self._conn, *kind)
        """Counters."""
        self.load_profile = GLoadProfile(self._conn, *kind)
        """Load Profile configurations."""

    async def delete(self):
        await P4G_DELETE(self._conn, *self.kind).set()
        self._observer.notify(idx_obs.IndexEvents.DEL, self)

    @classmethod
    async def _fetch(cls, conn: "itf.IConnection", module_id: int, port_id: int) -> List[int]:
        resp = await P4G_INDICES(conn, module_id, port_id).get()
        return list(resp.group_identifiers)

    @classmethod
    async def _new(cls: Type[CG], conn: "itf.IConnection", kind: "kind.IndicesKind", observer: "idx_obs.IndicesObserver") -> CG:
        await P4G_CREATE(conn, *kind).set()
        return cls(conn, kind, observer)



#: Available enums used by commands and server response status.

from .internals.core.commands.enums import *  # noqa: F403
from .internals.core.protocol.constants import CommandStatus

__all__ = (  # noqa: F405
    "AlgorithmMethod",
    "AlignLockStatus",
    "AnLtLogControl",
    "ApplicationLayerBehavior",
    "AutoNegFECOption",
    "AutoNegFECStatus",
    "AutoNegFECType",
    "AutoNegMode",
    "AutoNegStatus",
    "AutoNegTecAbility",
    "AutoOrManual",
    "BRRMode",
    "CaptureSize",
    "ChassisSessionType",
    "ChassisShutdownAction",
    "CommandStatus",
    "CongestionType",
    "CorruptionType",
    "CustomDefaultCommand",
    "CustomDefaultScope",
    "EmbedIP",
    "ErrorOnOff",
    "FaultSignaling",
    "FECMode",
    "FilterMode",
    "FilterType",
    "FilterUse",
    "FilterVlanType",
    "HasDemo",
    "HeaderLockStatus",
    "IGMPVersion",
    "ImpairmentLatencyMode",
    "ImpairmentTypeIndex",
    "InfiniteOrFinite",
    "InfoAction",
    "InjectErrorType",
    "IsEnabled",
    "IsOnline",
    "IsPermanent",
    "IsPresent",
    "IsValid",
    "L2PlusPresent",
    "L3PlusPresent",
    "L47IPVersion",
    "L47PortSpeed",
    "L47PortState",
    "L47ProtocolType",
    "L47TrafficState",
    "LatencyMode",
    "LatencyTypeCustomDist",
    "Layer1ConfigType",
    "Layer1LogType",
    "LengthCheckType",
    "LengthType",
    "LicenseSpeed",
    "LifecycleMode",
    "LinkState",
    "LinkTrainAlgorithm",
    "LinkTrainAnnounce",
    "LinkTrainCmd",
    "LinkTrainCmdFlags",
    "LinkTrainCmdResults",
    "LinkTrainCoeffs",
    "LinkTrainEncoding",
    "LinkTrainFrameLock",
    "LinkTrainingFailureType",
    "LinkTrainingInitCondition",
    "LinkTrainingMode",
    "LinkTrainingStatus",
    "LinkTrainingStatusMode",
    "LinkTrainPresets",
    "LocalFaultStatus",
    "LoopbackMode",
    "LoopBandwidth",
    "MDIXMode",
    "MediaCFPState",
    "MediaCFPType",
    "MediaConfigurationType",
    "ModifierAction",
    "MSSType",
    "MulticastExtOperation",
    "MulticastHeaderFormat",
    "MulticastOperation",
    "NRZPreset",
    "OnOff",
    "OnOffDefault",
    "OnOffWithSuppress",
    "PacketDetailSelection",
    "PacketType",
    "PAM4FrameSize",
    "PauseMode",
    "PayloadGenerationMethod",
    "PayloadMode",
    "PayloadType",
    "PFCMode",
    "PHYSignalStatus",
    "PolicerMode",
    "PortSpeedMode",
    "PPMSweepMode",
    "PPMSweepStatus",
    "PRBSInsertedType",
    "PRBSInvertState",
    "PRBSLockStatus",
    "PRBSOnOff",
    "PRBSPattern",
    "PRBSPolynomial",
    "PRBSStatisticsMode",
    "ProtocolOption",
    "ReconciliationSublayerSupport",
    "RemoteFaultStatus",
    "ReplayParserState",
    "ReplaySchedulingMode",
    "ReplaySyncBasedOn",
    "ReservedAction",
    "ReservedStatus",
    "ResourceAllocationMode",
    "RESTControlAction",
    "Role",
    "RTOType",
    "RXCState",
    "RXHState",
    "SerdesStatus",
    "ServiceStatus",
    "ShadowWorkingSelection",
    "SMAInputFunction",
    "SMAOutputFunction",
    "SMAStatus",
    "SourceType",
    "StartOrStop",
    "StartTrigger",
    "StopTrigger",
    "SyncStatus",
    "SystemUpdateStatus",
    "TimeKeeperLicenseError",
    "TimeKeeperLicenseFileState",
    "TimeKeeperLicenseType",
    "TimeKeeperServiceAction",
    "TimeKeeperServiceStatus",
    "TimeoutMode",
    "Timescale",
    "TimingSource",
    "TLSVersion",
    "TPLDMode",
    "TrafficEngine",
    "TrafficError",
    "TrafficOnOff",
    "TrafficScenario",
    "TSNConfigProfile",
    "TSNDeviationMode",
    "TSNHistogramSource",
    "TSNPortRole",
    "TSNStatisticsTypes",
    "TSNTimeSource",
    "TXClockSource",
    "TXClockStatus",
    "TXCState",
    "TXHState",
    "TXMode",
    "UpdateState",
    "WhoClose",
    "YesNo",
)



import typing
from xoa_driver import ports
from xoa_driver.internals.hli_v1 import revisions
from xoa_driver.internals.utils import ports_manager as pm

if typing.TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
    from .. import __interfaces as m_itf

from .module_l23_base import ModuleL23


@typing.final
@revisions.register_valkyrie_module(rev="Odin-40G-2S-2P")
class MOdin40G2S2P(ModuleL23):
    """Test module Odin-40G-2S-2P"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.POdin40G2S2P] = pm.PortsManager(
            conn=conn,
            ports_type=ports.POdin40G2S2P,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Odin-40G-2S-2P"""


@typing.final
@revisions.register_valkyrie_module(rev="Odin-40G-2S-2P-B")
class MOdin40G2S2PB(ModuleL23):
    """Test module Odin-40G-2S-2P-B"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.POdin40G2S2PB] = pm.PortsManager(
            conn=conn,
            ports_type=ports.POdin40G2S2PB,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Odin-40G-2S-2P-B"""



from typing import (
    TypeVar,
    Type,
    Protocol,
    ClassVar,
    Optional,
    Callable,
    Union,
)

XmpGenericType = TypeVar("XmpGenericType", bound="IXmpType")


class IXmpType(Protocol):
    size: ClassVar[int]
    __init__: Callable
    def __bytes__(self) -> bytes: ...  # noqa: E704
    def byte_length(self) -> int: ...  # noqa: E704
    @classmethod
    def from_bytes(cls: Type[XmpGenericType], data: bytes) -> XmpGenericType: ...  # noqa: E704


XmpGenericList = TypeVar("XmpGenericList", bound="IXmpDefaultList")


class IXmpDefaultList(Protocol):
    __init__: Callable
    size: ClassVar[int]
    element_type: ClassVar[Type[IXmpType]]
    fix_length: ClassVar[Optional[int]]
    stop_to_keep: ClassVar[Optional[int]]

    def byte_length(self) -> int: ...  # noqa: E704
    @classmethod
    def from_bytes(cls: Type[XmpGenericType], data: bytes) -> XmpGenericType: ...  # noqa: E704


XmpGenericField = TypeVar("XmpGenericField", bound=Union["IXmpDefaultList", "IXmpType"])



import functools
from typing import TYPE_CHECKING
from typing_extensions import Self

from xoa_driver.internals.core.commands import (
    P4_TRAFFIC,
    P4_STATE,
    P4_CAPABILITIES,
    P4_STATE_STATUS,
    P4_VLAN_OFFLOAD,
    P4_ARP_CONFIG,
    P4_NDP_CONFIG,
    P4_CAPTURE,
    P4_CAPTURE_GET_FIRST,
    P4_CAPTURE_GET_NEXT,
    P4_TX_PACKET_SIZE,
    P4_RX_PACKET_SIZE,
    P4_TX_MTU,
    P4_RX_MTU,
    P4_CLEAR,
    P4_SPEEDSELECTION,
    P4_MAX_PACKET_RATE,
    P4_PCI_INFO,
    P4_FW_VER,
    P4_DEV_NAME,
    P4_PORT_TYPE,
    P4_LICENSE_INFO,
    P4_APTITUDES,
)
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf

from xoa_driver.internals.hli_v2.ports import base_port
from xoa_driver.internals.state_storage import ports_state
from xoa_driver.internals.utils import attributes as utils
from xoa_driver.internals.utils.indices import index_manager as im
from xoa_driver.internals.hli_v2.indices.connection_group.cg import ConnectionGroupIdx

from .counters import PCounters
from .packet_engine import PacketEngine


class PCapture:
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.start = P4_CAPTURE(conn, module_id, port_id)
        self.get_first_frame = P4_CAPTURE_GET_FIRST(conn, module_id, port_id)
        self.get_next_frame = P4_CAPTURE_GET_NEXT(conn, module_id, port_id)


class PortL47(base_port.BasePort["ports_state.PortL47LocalState"]):
    """L47 Port"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        super().__init__(conn, module_id, port_id)

        self._local_states = ports_state.PortL47LocalState()

        self.traffic = P4_TRAFFIC(self._conn, module_id, port_id)
        """Representation of P4_TRAFFIC"""
        self.clear = P4_CLEAR(self._conn, module_id, port_id)
        """Representation of P4_CLEAR"""
        self.state = P4_STATE(self._conn, module_id, port_id)
        """Representation of P4_STATE"""
        self.type = P4_PORT_TYPE(self._conn, module_id, port_id)
        """Representation of P4_PORT_TYPE"""
        self.aptitudes = P4_APTITUDES(self._conn, module_id, port_id)
        """Representation of P4_APTITUDES"""
        self.last_state_status = P4_STATE_STATUS(self._conn, module_id, port_id)
        """Representation of P4_STATE_STATUS"""

        self.nic_name = P4_DEV_NAME(self._conn, module_id, port_id)
        """Representation of P4_DEV_NAME"""
        self.nic_firmware_version = P4_FW_VER(self._conn, module_id, port_id)
        """Representation of P4_FW_VER"""
        self.vlan_offload = P4_VLAN_OFFLOAD(self._conn, module_id, port_id)
        """Representation of P4_VLAN_OFFLOAD"""

        self.max_packet_rate = P4_MAX_PACKET_RATE(self._conn, module_id, port_id)
        """Representation of P4_MAX_PACKET_RATE"""
        self.capabilities = P4_CAPABILITIES(self._conn, module_id, port_id)
        """Representation of P4_CAPABILITIES"""
        self.speed_selection = P4_SPEEDSELECTION(self._conn, module_id, port_id)
        """Representation of P4_SPEEDSELECTION"""

        # region Not sure if this commands must be at this level
        self.tx_packet_size = P4_TX_PACKET_SIZE(self._conn, module_id, port_id)
        """Representation of P4_TX_PACKET_SIZE"""
        self.rx_packet_size = P4_RX_PACKET_SIZE(self._conn, module_id, port_id)
        """Representation of P4_RX_PACKET_SIZE"""
        self.tx_mtu = P4_TX_MTU(self._conn, module_id, port_id)
        """Representation of P4_TX_MTU"""
        self.rx_mtu = P4_RX_MTU(self._conn, module_id, port_id)
        """Representation of P4_RX_MTU"""
        # endregion

        self.pci_info = P4_PCI_INFO(self._conn, module_id, port_id)
        """Representation of P4_PCI_INFO"""
        self.license_info = P4_LICENSE_INFO(self._conn, module_id, port_id)
        """Representation of P4_LICENSE_INFO"""

        self.arp_config = P4_ARP_CONFIG(self._conn, module_id, port_id)
        """Representation of P4_ARP_CONFIG"""
        self.ndp_config = P4_NDP_CONFIG(self._conn, module_id, port_id)
        """Representation of P4_NDP_CONFIG"""

        self.capture = PCapture(self._conn, module_id, port_id)
        """L47 packet capture"""
        self.counters = PCounters(self._conn, module_id, port_id)
        """L47 counters"""
        self.packet_engine = PacketEngine(self._conn, module_id, port_id)
        """L47 packet engine"""

        self.connection_groups: "im.IndexManager[ConnectionGroupIdx]" = im.IndexManager(
            self._conn,
            ConnectionGroupIdx,
            module_id,
            port_id
        )
        """L47 connection group index manager."""

    @property
    def info(self) -> ports_state.PortL47LocalState:
        return self._local_states

    async def _setup(self) -> Self:
        await self._local_states.initiate(self)
        self._local_states.register_subscriptions(self)
        return self

    on_capabilities_change = functools.partialmethod(utils.on_event, P4_CAPABILITIES)
    """Register a callback to the event that the L47 port's capabilities change."""
    on_speed_selection_change = functools.partialmethod(utils.on_event, P4_SPEEDSELECTION)
    """Register a callback to the event that the L47 port's speed mode changes."""
    on_state_change = functools.partialmethod(utils.on_event, P4_STATE)
    """Register a callback to the event that the L47 port's state changes."""
    on_license_info_change = functools.partialmethod(utils.on_event, P4_LICENSE_INFO)
    """Register a callback to the event that the L47 port's license information changes."""



from typing import (
    Union,
    TYPE_CHECKING,
    Type,
)
from xoa_driver.internals.core.commands import (
    C_PORTCOUNTS,
    C_MULTIUSER,
    C_TRAFFIC,
    C_TRAFFICSYNC,
    C_VERSIONNO_MINOR,
    C_BUILDSTRING,
)
from xoa_driver.internals.utils.modules_manager import ModulesManager
from ._base_tester import BaseTester
from .genuine import management_interface as mi
from .genuine.l_23 import (
    upload_file,
    time_keeper,
    rest_api,
)
if TYPE_CHECKING:
    from xoa_driver import modules

from xoa_driver.internals.state_storage import testers_state
from xoa_driver.internals.hli_v1 import revisions
from xoa_driver.internals import exceptions


TypeL23Manager = ModulesManager[
    Union[
        "modules.GenericL23Module",
        "modules.ModuleChimera",
    ]
]


def get_module_type(revision: str) -> Type:
    module_type = dict(
        **revisions.VALKYRIE_MODULES,
        **revisions.CHIMERA_MODULES,
    ).get(revision)
    if not module_type:
        raise exceptions.WrongModuleError(
            revision,
            set.union(
                set(revisions.VALKYRIE_MODULES.keys()),
                set(revisions.CHIMERA_MODULES.keys()),
            )
        )
    return module_type


class L23Tester(BaseTester["testers_state.GenuineTesterLocalState"]):
    """
    This is a conceptual class of Xena Valkyrie Tester.
    It is essentially an extended :class:`BaseTester`.


    :param host: tester's address/hostname
    :type host: str
    :param username: username of the user
    :type username: str
    :param password: login password of the tester, defaults to "xena"
    :type password: str, optional
    :param port: the port number for connection establishment, default to 22606
    :type port: int, optional
    :param debug: `True` if debug log output from the tester is needed, and `False` otherwise
    :type debug: int, optional
    """
    
    def __init__(self, host: str, username: str, password: str = "xena", port: int = 22606, *, debug: bool = False) -> None:
        super().__init__(host=host, username=username, password=password, port=port, debug=debug)
        
        self._local_states = testers_state.GenuineTesterLocalState(host, port)
        
        self.management_interface = mi.ManagementInterface(self._conn)
        """
        The management interface address configuration includes IP address, DHCP settings, MAC address and hostname.
        
        :type: ManagementInterface
        """

        self.upload_file = upload_file.UploadFile(self._conn)
        """
        File upload functions of the tester.

        :type: UploadFile
        """

        self.rest_api_server = rest_api.RestApiServer(self._conn)
        """
        REST API server of the tester.

        :type: RestApiServer
        """

        self.time_keeper = time_keeper.TimeKeeper(self._conn)
        """
        TimeKeeper of the tester.

        :type: TimeKeeper
        """

        self.multiuser = C_MULTIUSER(self._conn)
        """
        Enable or disable the ability to control one resource from several different TCP
        connections.

        :type: C_MULTIUSER
        """

        self.traffic = C_TRAFFIC(self._conn)
        """
        Starts or stops the traffic on a number of ports on the chassis simultaneously.
        The ports are identified by pairs of integers (module port).

        :type: C_TRAFFIC
        """

        self.traffic_sync = C_TRAFFICSYNC(self._conn)
        """
        This can be used to start traffic simultaneously on multiple chassis. The ports are identified by pairs of integers (module port).

        :type: C_TRAFFICSYNC
        """
        
        self.version_no_minor = C_VERSIONNO_MINOR(self._conn)
        """
        Get the minor version number of the tester firmware.

        :type: C_VERSIONNO_MINOR
        """

        self.build_string = C_BUILDSTRING(self._conn)
        """
        Identify the hostname of the PC that builds the xenaserver. It uniquely identifies the build of a xenaserver.

        :type: C_BUILDSTRING
        """

        self.modules: TypeL23Manager = ModulesManager(self._conn, get_module_type)
        """
        Module Index Manager of the L23 tester.

        :type: ModulesManager
        """
    
    @property
    def info(self) -> testers_state.GenuineTesterLocalState:
        """Return tester's local state

        :return: tester's local state
        :rtype: GenuineTesterLocalState
        """
        return self._local_states
    
    async def _setup(self):
        await super()._setup()
        await self._local_states.initiate(self)
        self._local_states.register_subscriptions(self)
        
        ft_pc = await C_PORTCOUNTS(self._conn).get()
        port_counts = ft_pc.port_counts
        await self.modules.fill_l23(port_counts)
        return self

